# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.
#
#
# Install this as hyperspy/io_plugins/phenom.py
#
# If you get the error "cannot decompress LZW", run
# pip install imagecodecs
#
# Edit hyperspy/io_plugins/__init__.py and add phenom to both the
# list of imports and io_plugins
#
# You should now be able to load Phenom EID .elid files.
#
# This reader supports the ELID file format used in Phenom ProSuite
# Element Identification version 3.8.0 and later. You can convert
# older ELID files by loading the file into a recent Element
# Identification release and then save the ELID file into the newer
# file format.

import bz2
import copy
import io
import math
import os
import struct
import xml.etree.ElementTree as ET
from datetime import datetime

import numpy as np
import tifffile
from dateutil import tz

from rsciio._docstrings import FILENAME_DOC, LAZY_UNSUPPORTED_DOC, RETURNS_DOC


def element_symbol(z):
    elements = [
        "",
        "H",
        "He",
        "Li",
        "Be",
        "B",
        "C",
        "N",
        "O",
        "F",
        "Ne",
        "Na",
        "Mg",
        "Al",
        "Si",
        "P",
        "S",
        "Cl",
        "Ar",
        "K",
        "Ca",
        "Sc",
        "Ti",
        "V",
        "Cr",
        "Mn",
        "Fe",
        "Co",
        "Ni",
        "Cu",
        "Zn",
        "Ga",
        "Ge",
        "As",
        "Se",
        "Br",
        "Kr",
        "Rb",
        "Sr",
        "Y",
        "Zr",
        "Nb",
        "Mo",
        "Tc",
        "Ru",
        "Rh",
        "Pd",
        "Ag",
        "Cd",
        "In",
        "Sn",
        "Sb",
        "Te",  # codespell:ignore
        "I",
        "Xe",
        "Cs",
        "Ba",
        "La",
        "Ce",
        "Pr",
        "Nd",  # codespell:ignore
        "Pm",
        "Sm",
        "Eu",
        "Gd",
        "Tb",
        "Dy",
        "Ho",
        "Er",
        "Tm",
        "Yb",
        "Lu",
        "Hf",
        "Ta",
        "W",
        "Re",
        "Os",
        "Ir",
        "Pt",
        "Au",
        "Hg",
        "Tl",
        "Pb",
        "Bi",
        "Po",
        "At",
        "Rn",
        "Fr",
        "Ra",
        "Ac",
        "Th",
        "Pa",
        "U",
        "Np",
        "Pu",
        "Am",
        "Cm",
        "Bk",
        "Cf",
        "Es",
        "Fm",
        "Md",
        "No",
        "Lr",
        "Rf",
        "Db",
        "Sg",
        "Bh",
        "Hs",
        "Mt",
        "Ds",
        "Rg",
        "Cn",
        "Nh",
        "Fl",
        "Mc",
        "Lv",
        "Ts",
        "Og",
    ]
    if z < 1 or z >= len(elements):
        raise Exception("Invalid atomic number")
    return elements[z]


def family_symbol(i):
    families = ["", "K", "L", "M", "N", "O", "P"]
    if i < 1 or i >= len(families):
        raise Exception("Invalid atomic number")
    return families[i]


def IsGZip(pathname):
    with open(pathname, "rb") as f:
        (magic,) = struct.unpack("2s", f.read(2))
    return magic == b"\x1f\x8b"


def IsBZip2(pathname):
    with open(pathname, "rb") as f:
        (magic, _, bytes) = struct.unpack("2s2s6s", f.read(10))
    return magic == b"BZ" and bytes == b"\x31\x41\x59\x26\x53\x59"


class ElidReader:
    def __init__(self, pathname, block_size=1024 * 1024):
        if IsGZip(pathname):
            raise Exception("pre EID 3.8 files are not supported")
        if not IsBZip2(pathname):
            raise Exception("not an ELID file")

        self._pathname = pathname
        with open(pathname, "rb") as self._file:
            self._decompressor = bz2.BZ2Decompressor()
            self._block_size = block_size
            (id, version) = struct.unpack("<4si", self._read(8))
            if id != b"EID2":
                raise Exception("Not an ELID file.")
            if version > 4:
                raise Exception(f"unsupported ELID format {version}.")
            self._version = version
            self.dictionaries = self._read_Project()

    def _read(self, size=1):
        data = self._decompressor.decompress(b"", size)
        while self._decompressor.needs_input:
            data += self._decompressor.decompress(
                self._file.read(self._block_size), size - len(data)
            )
        return data

    def _read_bool(self):
        return struct.unpack("?", self._read(1))[0]

    def _read_uint8(self):
        return struct.unpack("B", self._read(1))[0]

    def _read_int32(self):
        return struct.unpack("<i", self._read(4))[0]

    def _read_uint32(self):
        return struct.unpack("<I", self._read(4))[0]

    def _read_string(self):
        n = self._read_uint32()
        return self._read(n).decode("utf-8")

    def _get_unit_factor(self, unit):
        if len(unit) < 2:
            return 1
        elif unit[0] == "M":
            return 1e6
        elif unit[0] == "k":
            return 1e3
        elif unit[0] == "m":
            return 1e-3
        elif unit[0] == "u" or unit[0] == "µ":
            return 1e-6
        elif unit[0] == "n":
            return 1e-9
        elif unit[0] == "p":
            return 1e-12
        else:
            raise Exception("Unknown unit: " + unit)

    def _get_value_with_unit(self, item):
        if isinstance(item, dict):
            return float(item["value"]) * self._get_unit_factor(item["unit"])
        else:
            return float(item)

    def _read_tiff(self):
        def xml_element_to_dict(element):
            dict = {}
            if len(element) == 0:
                if len(element.items()) > 0:
                    dict[element.tag] = {"value": element.text}
                    for attrib, value in element.items():
                        dict[element.tag].update({attrib: value})
                else:
                    dict[element.tag] = element.text
            else:
                dict[element.tag] = {}
                for child in element:
                    dict[element.tag].update(xml_element_to_dict(child))
            return dict

        def make_metadata_dict(xml):
            dict = xml_element_to_dict(ET.fromstring(xml))
            return dict["FeiImage"] if dict else {}

        n = self._read_uint32()
        if n == 0:
            return (None, None)
        bytes = io.BytesIO(self._read(n))
        with tifffile.TiffFile(bytes) as tiff:
            data = tiff.asarray()
            if len(data.shape) > 2:
                # HyperSpy uses struct arrays to store RGB data
                from rsciio.utils import rgb

                data = rgb.regular_array2rgbx(data)

            tags = tiff.pages[0].tags
            if "FEI_TITAN" in tags:
                metadata = make_metadata_dict(tags["FEI_TITAN"].value)
                metadata["acquisition"]["scan"]["fieldSize"] = max(
                    self._get_value_with_unit(metadata["pixelHeight"]) * data.shape[0],
                    self._get_value_with_unit(metadata["pixelWidth"]) * data.shape[1],
                )
            else:
                metadata = {}
        return (metadata, data)

    def _read_int32s(self):
        n = self._read_uint32()
        return [self._read_int32() for _ in range(n)]

    def _read_float64(self):
        return struct.unpack("<d", self._read(8))[0]

    def _read_float64s(self):
        n = self._read_uint32()
        return [self._read_float64() for _ in range(n)]

    def _read_varuint32(self):
        value = 0
        shift = 0
        while True:
            b = self._read_uint8()
            value = value | ((b & 127) << shift)
            if (b & 128) == 0:
                break
            shift += 7
        return value

    def _read_spectrum(self):
        offset = self._read_float64()
        dispersion = self._read_float64()
        n = self._read_uint32()
        return (offset, dispersion, [self._read_varuint32() for _ in range(n)])

    def _read_uint8s(self):
        n = self._read_uint32()
        return [self._read_uint8() for _ in range(n)]

    def _read_oxide(self):
        element = self._read_uint8()
        num_element = self._read_uint8()
        num_oxygen = self._read_uint8()
        oxide = element_symbol(element)
        if num_element > 1:
            oxide += str(num_element)
        oxide += "O"
        if num_oxygen > 1:
            oxide += str(num_oxygen)
        return oxide

    def _read_oxides(self):
        n = self._read_uint32()
        return [self._read_oxide() for _ in range(n)]

    def _read_element_family(self):
        element = element_symbol(self._read_uint8())
        family = family_symbol(self._read_uint8())
        return (element, family)

    def _read_element_families(self):
        n = self._read_uint32()
        return [self._read_element_family() for _ in range(n)]

    def _read_drift_correction(self):
        dc = self._read_uint8()
        if dc == 1:
            return "on"
        elif dc == 2:
            return "off"
        else:
            return "unknown"

    def _read_detector_type(self):
        dt = self._read_uint8()
        if dt == 0:
            return "Empty"
        elif dt == 1:
            return "FastSDD_C2"
        elif dt == 2:
            return "FastSDD_C5"
        elif dt == 3:
            return "FastSDD_WLS"
        else:
            return "Unknown"

    def _read_spectrum_correction(self):
        sc = self._read_uint8()
        if sc == 1:
            return "linearized"
        elif sc == 2:
            return "raw"
        else:
            return "unknown"

    def _read_element_collection(self):
        return [element_symbol(z) for z in self._read_uint8s()]

    def _read_eds_metadata(self, om):
        metadata = {}
        metadata["high_tension"] = self._read_float64()
        detector_elevation = self._read_float64()
        if self._version == 0:
            detector_elevation = math.radians(detector_elevation)
        metadata["detector_elevation"] = detector_elevation
        metadata["detector_azimuth"] = self._read_float64()
        metadata["live_time"] = self._read_float64()
        metadata["real_time"] = self._read_float64()
        metadata["slow_peaking_time"] = self._read_float64()
        metadata["fast_peaking_time"] = self._read_float64()
        metadata["detector_resolution"] = self._read_float64()
        metadata["instrument_id"] = self._read_string()
        if self._version == 0 and "workingDistance" in om:
            metadata["working_distance"] = self._get_value_with_unit(
                om["workingDistance"]
            )
            metadata["slow_peaking_time"] = (
                11.2e-6 if float(om["acquisition"]["scan"]["spotSize"]) < 4.5 else 2e-6
            )
            metadata["fast_peaking_time"] = 100e-9
            metadata["detector_surface_area"] = 25e-6
        elif self._version > 0:
            metadata["optical_working_distance"] = self._read_float64()
            metadata["working_distance"] = self._read_float64()
            metadata["detector_surface_area"] = self._read_float64()
            metadata["detector_distance"] = self._read_float64()
            metadata["sample_tilt_angle"] = self._read_float64()
        if self._version >= 3:
            metadata["ccorrection"] = self._read_float64()
            metadata["detector_type"] = self._read_detector_type()
            metadata["spectrum_correction"] = self._read_spectrum_correction()
        else:
            metadata["ccorrection"] = 0
            metadata["detector_type"] = "FastSDD_C2"
            metadata["spectrum_correction"] = "unknown"
        return metadata

    def _read_CommonAnalysis(self, am):
        (metadata, cutout) = self._read_tiff()
        sum_spectrum = self._read_spectrum()
        eds_metadata = self._read_eds_metadata(am)
        eds_metadata["offset"] = sum_spectrum[0]
        eds_metadata["dispersion"] = sum_spectrum[1]
        data = sum_spectrum[2]
        eds_metadata["included_elements"] = self._read_element_collection()
        eds_metadata["excluded_elements"] = self._read_element_collection()
        eds_metadata["background_fit_bins"] = self._read_int32s()
        eds_metadata["selected_oxides"] = self._read_oxides()
        eds_metadata["auto_id"] = self._read_bool()
        eds_metadata["order_nr"] = self._read_int32()
        eds_metadata["family_overrides"] = self._read_element_families()
        if self._version >= 2:
            eds_metadata["drift_correction"] = self._read_drift_correction()
        else:
            eds_metadata["drift_correction"] = "unknown"
        if self._version >= 4:
            eds_metadata["ignored_elements"] = self._read_element_collection()
        else:
            eds_metadata["ignored_elements"] = []
        if metadata:
            metadata["acquisition"]["scan"]["detectors"]["EDS"] = eds_metadata
        else:
            metadata = {}
            metadata.update({"acquisition": {"scan": {"detectors": {}}}})
            metadata["acquisition"]["scan"]["detectors"]["EDS"] = eds_metadata
        return (metadata, data)

    def _make_metadata_dict(self, signal_type=None, title="", datetime=None):
        metadata_dict = {
            "General": {
                "original_filename": os.path.split(self._pathname)[1],
                "title": title,
            }
        }
        if signal_type:
            metadata_dict["Signal"] = {"signal_type": signal_type}
        if datetime:
            metadata_dict["General"].update(
                {
                    "date": datetime[0],
                    "time": datetime[1],
                    "time_zone": self._get_local_time_zone(),
                }
            )
        return metadata_dict

    def _get_local_time_zone(self):
        return tz.tzlocal().tzname(datetime.today())

    def _make_mapping(self):
        return {
            "acquisition.scan.detectors.EDS.detector_azimuth": (
                "Acquisition_instrument.SEM.Detector.EDS.azimuth_angle",
                lambda x: math.degrees(float(x)),
            ),
            "acquisition.scan.detectors.EDS.detector_elevation": (
                "Acquisition_instrument.SEM.Detector.EDS.elevation_angle",
                lambda x: math.degrees(float(x)),
            ),
            "acquisition.scan.detectors.EDS.detector_resolution": (
                "Acquisition_instrument.SEM.Detector.EDS.energy_resolution_MnKa",
                float,
            ),
            "acquisition.scan.detectors.EDS.live_time": (
                "Acquisition_instrument.SEM.Detector.EDS.live_time",
                float,
            ),
            "acquisition.scan.detectors.EDS.real_time": (
                "Acquisition_instrument.SEM.Detector.EDS.real_time",
                float,
            ),
            "acquisition.scan.detectors.EDS.high_tension": (
                "Acquisition_instrument.SEM.beam_energy",
                lambda x: float(x) / 1e3,
            ),
            "acquisition.scan.highVoltage.value": (
                "Acquisition_instrument.SEM.beam_energy",
                lambda x: -float(x),
            ),
            "instrument.uniqueID": (
                "Acquisition_instrument.SEM.microscope",
                lambda x: x,
            ),
            "samplePosition.x": (
                "Acquisition_instrument.SEM.Stage.x",
                lambda x: float(x) / 1e-3,
            ),
            "samplePosition.y": (
                "Acquisition_instrument.SEM.Stage.y",
                lambda x: float(x) / 1e-3,
            ),
            "acquisition.scan.detectors.EDS.sample_tilt_angle": (
                "Acquisition_instrument.SEM.Stage.tilt_alpha",
                lambda x: math.degrees(float(x)),
            ),
            "acquisition.scan.detectors.EDS.working_distance": (
                "Acquisition_instrument.SEM.working_distance",
                lambda x: float(x) / 1e-3,
            ),
        }

    def _make_spot_spectrum_dict(self, om, offset, dispersion, data, title):
        axes = [
            {
                "name": "Energy",
                "offset": offset / 1e3,
                "scale": dispersion / 1e3,
                "size": len(data),
                "units": "keV",
                "navigate": False,
            }
        ]
        dict = {
            "data": data,
            "axes": axes,
            "metadata": self._make_metadata_dict(
                "EDS_SEM", title, self._get_datetime(om)
            ),
            "original_metadata": om,
            "mapping": self._make_mapping(),
        }
        return dict

    def _get_datetime(self, metadata):
        if "time" in metadata:
            return metadata["time"].split("T")
        else:
            return None

    def _make_line_spectrum_dict(self, om, offset, dispersion, data, title):
        axes = [
            {
                "index_in_array": 0,
                "name": "i",
                "offset": 0,
                "scale": 1,
                "size": data.shape[0],
                "units": "points",
                "navigate": True,
            },
            {
                "index_in_array": 1,
                "name": "X-ray energy",
                "offset": offset / 1e3,
                "scale": dispersion / 1e3,
                "size": data.shape[1],
                "units": "keV",
                "navigate": False,
            },
        ]
        dict = {
            "data": data,
            "axes": axes,
            "metadata": self._make_metadata_dict(
                "EDS_SEM", title, self._get_datetime(om)
            ),
            "original_metadata": om,
            "mapping": self._make_mapping(),
        }
        return dict

    def _get_unit(self, value):
        if value > 1:
            return (1, "")
        elif value > 1e-3:
            return (1e-3, "m")
        elif value > 1e-6:
            return (1e-6, "µ")
        elif value > 1e-9:
            return (1e-9, "n")
        else:
            return (1, "")

    def _make_map_spectrum_dict(self, om, offset, dispersion, data, title):
        size = om["acquisition"]["scan"]["fieldSize"] * float(
            om["acquisition"]["scan"]["scanScale"]
        )
        (scale, prefix) = self._get_unit(size)
        unit = prefix + "m"
        size = size / scale
        axes = [
            {
                "index_in_array": 0,
                "name": "y",
                "offset": 0,
                "scale": size / data.shape[0],
                "size": data.shape[0],
                "units": unit,
                "navigate": True,
            },
            {
                "index_in_array": 1,
                "name": "x",
                "offset": 0,
                "scale": size / data.shape[1],
                "size": data.shape[1],
                "units": unit,
                "navigate": True,
            },
            {
                "index_in_array": 2,
                "name": "X-ray energy",
                "offset": offset / 1e3,
                "scale": dispersion / 1e3,
                "size": data.shape[2],
                "units": "keV",
                "navigate": False,
            },
        ]
        dict = {
            "data": data,
            "axes": axes,
            "metadata": self._make_metadata_dict(
                "EDS_SEM", title, self._get_datetime(om)
            ),
            "original_metadata": om,
            "mapping": self._make_mapping(),
        }
        return dict

    def _make_image_dict(self, om, data, title):
        if om:
            (scale, prefix) = self._get_unit(
                0.2 * om["acquisition"]["scan"]["fieldSize"]
            )
            scale_x = self._get_value_with_unit(om["pixelWidth"]) / scale
            scale_y = self._get_value_with_unit(om["pixelHeight"]) / scale
            unit = prefix + "m"
        else:
            scale_x = 1
            scale_y = 1
            unit = "points"
        axes = [
            {
                "index_in_array": 0,
                "name": "y",
                "offset": 0,
                "scale": scale_y,
                "size": data.shape[0],
                "units": unit,
                "navigate": True,
            },
            {
                "index_in_array": 1,
                "name": "x",
                "offset": 0,
                "scale": scale_x,
                "size": data.shape[1],
                "units": unit,
                "navigate": True,
            },
        ]
        dict = {
            "data": data,
            "axes": axes,
            "metadata": self._make_metadata_dict("", title, self._get_datetime(om)),
            "original_metadata": om,
            "mapping": self._make_mapping(),
        }
        return dict

    def _read_MsaAnalysis(self, label, am):
        (om, sum_spectrum) = self._read_CommonAnalysis(am)
        original_metadata = copy.deepcopy(am)
        original_metadata.update(om)
        return self._make_spot_spectrum_dict(
            original_metadata,
            om["acquisition"]["scan"]["detectors"]["EDS"]["offset"],
            om["acquisition"]["scan"]["detectors"]["EDS"]["dispersion"],
            np.array(sum_spectrum),
            "{}, MSA {}".format(
                label, om["acquisition"]["scan"]["detectors"]["EDS"]["order_nr"]
            ),
        )

    def _read_SpotAnalysis(self, label, am):
        (om, sum_spectrum) = self._read_CommonAnalysis(am)
        x = self._read_float64()
        y = self._read_float64()
        original_metadata = copy.deepcopy(am)
        original_metadata["acquisition"]["scan"]["detectors"]["EDS"] = om[
            "acquisition"
        ]["scan"]["detectors"]["EDS"]
        original_metadata["acquisition"]["scan"]["detectors"]["EDS"]["position"] = {
            "x": x,
            "y": y,
        }
        return self._make_spot_spectrum_dict(
            original_metadata,
            om["acquisition"]["scan"]["detectors"]["EDS"]["offset"],
            om["acquisition"]["scan"]["detectors"]["EDS"]["dispersion"],
            np.array(sum_spectrum),
            "{}, Spot {}".format(
                label, om["acquisition"]["scan"]["detectors"]["EDS"]["order_nr"]
            ),
        )

    def _read_LineScanAnalysis(self, label, am):
        (om, sum_spectrum) = self._read_CommonAnalysis(am)
        x1 = self._read_float64()
        y1 = self._read_float64()
        x2 = self._read_float64()
        y2 = self._read_float64()
        size = self._read_uint32()
        bins = self._read_uint32()
        offset = self._read_float64()
        dispersion = self._read_float64()
        eds_metadata = self._read_eds_metadata(am)
        eds_metadata["live_time"] = om["acquisition"]["scan"]["detectors"]["EDS"][
            "live_time"
        ]
        eds_metadata["real_time"] = om["acquisition"]["scan"]["detectors"]["EDS"][
            "real_time"
        ]
        eds_metadata["begin"] = {"x": x1, "y": y1}
        eds_metadata["end"] = {"x": x2, "y": y2}
        eds_metadata["offset"] = offset
        eds_metadata["dispersion"] = dispersion
        has_variable_real_time = self._read_bool()
        has_variable_live_time = self._read_bool()
        data = np.empty([size, bins], dtype=np.uint32)
        for i in range(size):
            for bin in range(bins):
                data[i, bin] = self._read_varuint32()
        if has_variable_real_time:
            eds_metadata["real_time_values"] = [
                self._read_float64() for _ in range(size)
            ]
        else:
            eds_metadata["real_time_values"] = [self._read_float64()] * size
        if has_variable_live_time:
            eds_metadata["live_time_values"] = [
                self._read_float64() for _ in range(size)
            ]
        else:
            eds_metadata["live_time_values"] = [self._read_float64()] * size
        eds_metadata["high_accuracy_quantification"] = self._read_bool()
        original_metadata = copy.deepcopy(am)
        original_metadata["acquisition"]["scan"]["detectors"]["EDS"] = eds_metadata
        return self._make_line_spectrum_dict(
            original_metadata,
            om["acquisition"]["scan"]["detectors"]["EDS"]["offset"],
            om["acquisition"]["scan"]["detectors"]["EDS"]["dispersion"],
            data,
            "{}, Line {}".format(
                label, om["acquisition"]["scan"]["detectors"]["EDS"]["order_nr"]
            ),
        )

    def _read_MapAnalysis(self, label, am):
        (om, sum_spectrum) = self._read_CommonAnalysis(am)
        # These metadata are currently not used but we still need to
        # read these to advance the position in the file
        # use placeholder for readability
        _ = self._read_float64()  # left
        _ = self._read_float64()  # top
        _ = self._read_float64()  # right
        _ = self._read_float64()  # bottom
        _ = self._read_float64s()  # color_intensities
        width = self._read_uint32()
        height = self._read_uint32()
        bins = self._read_uint32()
        _ = self._read_float64()  # offset
        _ = self._read_float64()  # dispersion
        original_metadata = copy.deepcopy(am)
        eds_metadata = self._read_eds_metadata(am)
        eds_metadata["live_time"] = om["acquisition"]["scan"]["detectors"]["EDS"][
            "live_time"
        ]
        eds_metadata["real_time"] = om["acquisition"]["scan"]["detectors"]["EDS"][
            "real_time"
        ]
        original_metadata["acquisition"]["scan"]["detectors"]["EDS"] = eds_metadata
        has_variable_real_time = self._read_bool()
        has_variable_live_time = self._read_bool()
        data = np.empty([height, width, bins], dtype=np.uint32)
        for y in range(height):
            for x in range(width):
                for bin in range(bins):
                    data[y, x, bin] = self._read_varuint32()
        if has_variable_real_time:
            real_time_values = np.empty([height, width], dtype=float)
            for y in range(height):
                for x in range(width):
                    real_time_values[y, x] = self._read_float64()
            eds_metadata["real_time_values"] = real_time_values
        else:
            eds_metadata["real_time_values"] = np.full(
                [height, width], self._read_float64()
            )
        if has_variable_live_time:
            live_time_values = np.empty([height, width], dtype=float)
            for y in range(height):
                for x in range(width):
                    live_time_values[y, x] = self._read_float64()
            eds_metadata["live_time_values"] = live_time_values
        else:
            eds_metadata["live_time_values"] = np.full(
                [height, width], self._read_float64()
            )
        return self._make_map_spectrum_dict(
            original_metadata,
            om["acquisition"]["scan"]["detectors"]["EDS"]["offset"],
            om["acquisition"]["scan"]["detectors"]["EDS"]["dispersion"],
            data,
            "{}, Map {}".format(
                label, om["acquisition"]["scan"]["detectors"]["EDS"]["order_nr"]
            ),
        )

    def _read_DifferenceAnalysis(self, label, am):
        (om, sum_spectrum) = self._read_CommonAnalysis(am)
        minuend = self._read_uint32()
        subtrahend = self._read_uint32()
        original_metadata = copy.deepcopy(am)
        original_metadata["acquisition"]["scan"]["detectors"]["EDS"] = om[
            "acquisition"
        ]["scan"]["detectors"]["EDS"]
        return self._make_spot_spectrum_dict(
            original_metadata,
            om["acquisition"]["scan"]["detectors"]["EDS"]["offset"],
            om["acquisition"]["scan"]["detectors"]["EDS"]["dispersion"],
            np.array(sum_spectrum),
            "{}, Difference {} - {}".format(label, minuend, subtrahend),
        )

    def _read_RegionAnalysis(self, label, am):
        (om, sum_spectrum) = self._read_CommonAnalysis(am)
        left = self._read_float64()
        top = self._read_float64()
        right = self._read_float64()
        bottom = self._read_float64()
        original_metadata = copy.deepcopy(am)
        original_metadata["acquisition"]["scan"]["detectors"]["EDS"] = om[
            "acquisition"
        ]["scan"]["detectors"]["EDS"]
        original_metadata["acquisition"]["scan"]["detectors"]["EDS"]["rectangle"] = {
            "left": left,
            "top": top,
            "right": right,
            "bottom": bottom,
        }
        return self._make_spot_spectrum_dict(
            original_metadata,
            om["acquisition"]["scan"]["detectors"]["EDS"]["offset"],
            om["acquisition"]["scan"]["detectors"]["EDS"]["dispersion"],
            np.array(sum_spectrum),
            "{}, Region {}".format(
                label, om["acquisition"]["scan"]["detectors"]["EDS"]["order_nr"]
            ),
        )

    def _read_ConstructiveAnalysisSource(self):
        analysis_index = self._read_uint32()
        weight_factor = self._read_float64()
        return (analysis_index, weight_factor)

    def _read_ConstructiveAnalysisSources(self):
        n = self._read_uint32()
        return [self._read_ConstructiveAnalysisSource() for _ in range(n)]

    def _read_ConstructiveAnalysis(self, label, am):
        self._read_CommonAnalysis(am)
        # These metadata are currently not used but we still need to
        # read these to advance the position in the file
        # use placeholder for readability
        _ = self._read_string()  # description
        _ = self._read_ConstructiveAnalysisSources()  # sources

    def _read_ConstructiveAnalyses(self):
        return self._read_Analyses("", {})

    def _read_Analysis(self, label, am):
        type = self._read_uint8()
        if type == 1:
            return self._read_MsaAnalysis(label, am)
        elif type == 2:
            return self._read_SpotAnalysis(label, am)
        elif type == 3:
            return self._read_LineScanAnalysis(label, am)
        elif type == 4:
            return self._read_MapAnalysis(label, am)
        elif type == 5:
            return self._read_DifferenceAnalysis(label, am)
        elif type == 6:
            return self._read_RegionAnalysis(label, am)
        elif type == 7:
            return self._read_ConstructiveAnalysis(label, am)
        else:
            raise Exception("Unknown Analysis type")

    def _read_Analyses(self, label, metadata):
        n = self._read_uint32()
        return [self._read_Analysis(label, metadata) for _ in range(n)]

    def _read_Image(self):
        tiff = self._read_tiff()
        label = self._read_string()
        dictionaries = []
        if tiff:
            dictionaries.append(self._make_image_dict(tiff[0], tiff[1], label))
        dictionaries.extend(self._read_Analyses(label, tiff[0]))
        return dictionaries

    def _read_Images(self):
        n = self._read_uint32()
        dictionaries = []
        for _ in range(n):
            dictionaries.extend(self._read_Image())
        return dictionaries

    def _read_Project(self):
        dictionaries = self._read_Images()
        if self._version >= 1:
            self._read_Analyses("", {})
            self._read_ConstructiveAnalyses()
        return [dict for dict in dictionaries if dict]


def file_reader(filename, lazy=False):
    """
    Read a Phenom ``.elid`` file from the software Element Identification (>v3.8.0)
    used by the Thermo Fisher Scientific Phenom desktop SEMs.

    Parameters
    ----------
    %s
    %s

    %s
    """
    if lazy is not False:
        raise NotImplementedError("Lazy loading is not supported.")

    reader = ElidReader(filename)
    return reader.dictionaries


file_reader.__doc__ %= (FILENAME_DOC, LAZY_UNSUPPORTED_DOC, RETURNS_DOC)



import csv
import logging
import os

import numpy as np

from rsciio._docstrings import FILENAME_DOC, LAZY_UNSUPPORTED_DOC, RETURNS_DOC

_logger = logging.getLogger(__name__)


# At some point, if there is another reader, which also use csv file, it will
# be necessary to mention the other reader in this message (and to add an
# argument in the load function to specify the correct reader)
invalid_file_error = (
    "The csv reader can't import the file, please"
    " make sure, that this is a valid Impulse log file."
)
invalid_filenaming_error = {
    "The filename does not match Impulse naming, please"
    " make sure that the filenames for the logfile and metadata file are unchanged."
}


def file_reader(filename, lazy=False):
    """
    Read a DENSsolutions Impulse logfile.

    Parameters
    ----------
    %s
    %s

    %s
    """
    if lazy is not False:
        raise NotImplementedError("Lazy loading is not supported.")

    csv_file = ImpulseCSV(filename)

    return _impulseCSV_log_reader(csv_file)


file_reader.__doc__ %= (FILENAME_DOC, LAZY_UNSUPPORTED_DOC, RETURNS_DOC)


def _impulseCSV_log_reader(csv_file):
    csvs = []
    for key in csv_file.logged_quantity_name_list:
        csvs.append(csv_file.get_dictionary(key))
    return csvs


class ImpulseCSV:
    def __init__(self, filename):
        self.filename = filename
        self._parse_header()
        self._read_data()

    def _parse_header(self):
        with open(self.filename, "r") as f:
            s = f.readline()
            self.column_names = s.strip().split(",")
            if not self._is_impulse_csv_file():
                raise IOError(invalid_file_error)
            self._read_metadatafile()
        self.logged_quantity_name_list = self.column_names[2:]

    def _is_impulse_csv_file(self):
        return "TimeStamp" in self.column_names and len(self.column_names) >= 3

    def get_dictionary(self, quantity):
        return {
            "data": self._data_dictionary[quantity],
            "axes": self._get_axes(),
            "metadata": self._get_metadata(quantity),
            "original_metadata": {"Impulse_header": self.original_metadata},
        }

    def _get_metadata(self, quantity):
        return {
            "General": {
                "original_filename": os.path.split(self.filename)[1],
                "title": "%s" % quantity,
                "date": self.original_metadata["Experiment_date"],
                "time": self.original_metadata["Experiment_time"],
            },
            "Signal": {
                "quantity": self._parse_quantity_units(quantity),
            },
        }

    def _parse_quantity_units(self, quantity):
        quantity_split = quantity.strip().split(" ")
        if (
            len(quantity_split) > 1
            and quantity_split[-1][0] == "("
            and quantity_split[-1][-1] == ")"
        ):
            return quantity_split[-1].replace("(", "").replace(")", "")
        else:
            return ""

    def _read_data(self):
        names = [
            name.replace(" ", "_")
            .replace("°C", "C")
            .replace("#", "No")
            .replace("(", "")
            .replace(")", "")
            .replace("/", "_")
            .replace("%", "Perc")
            for name in self.column_names
        ]
        data = np.genfromtxt(
            self.filename,
            delimiter=",",
            dtype=None,
            names=names,
            skip_header=1,
            encoding="latin1",
        )
        self._data_dictionary = dict()
        for i, (name, name_dtype) in enumerate(zip(self.column_names, names)):
            if name == "Experiment time":
                self.time_axis = data[name_dtype]
            elif name == "MixValve":
                mixvalvedatachanged = data[name_dtype]
                for index, item in enumerate(data[name_dtype]):
                    mixvalvedatachanged[index] = (
                        int(int(item.split(";")[0]) + 2) * 100
                        + (int(item.split(";")[1]) + 2) * 10
                        + (int(item.split(";")[2]) + 2)
                    )
                mixvalvedatachangedint = np.array(mixvalvedatachanged, dtype=np.int32)
                self._data_dictionary[name] = mixvalvedatachangedint
            else:
                self._data_dictionary[name] = data[name_dtype]

    def _read_metadatafile(self):
        # Locate the experiment metadata file
        self.original_metadata = {}
        notes = []
        notes_section = False

        if "_Synchronized data" in str(self.filename) or "raw" in str(
            self.filename
        ):  # Check if Impulse filename formatting is intact
            metadata_file = (
                "_".join(str(self.filename).split("_")[:-1]) + "_Metadata.log"
            ).replace("\\", "/")
            if os.path.isfile(metadata_file):
                with open(metadata_file, newline="") as csvfile:
                    metadata_file_reader = csv.reader(csvfile, delimiter=",")
                    for row in metadata_file_reader:
                        if notes_section:
                            notes.append(row[0])
                        elif row[0] == "Live notes":
                            notes_section = True
                            notes = [row[1].strip()]
                        else:
                            self.original_metadata[row[0].replace(" ", "_")] = row[
                                1
                            ].strip()
                    self.original_metadata["Notes"] = notes

            else:
                _logger.warning("No metadata file found in folder")
        else:
            raise IOError(invalid_filenaming_error)

    def _get_axes(self):
        return [
            {
                "size": self.time_axis.shape[0],
                "index_in_array": 0,
                "name": "Time",
                "scale": np.diff(self.time_axis[1:-1]).mean(),
                "offset": 0,
                "units": "Seconds",
                "navigate": False,
            }
        ]



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

# The details of the format were taken from
# https://www.biochem.mpg.de/doc_tom/TOM_Release_2008/IOfun/tom_mrcread.html
# and https://ami.scripps.edu/software/mrctools/mrc_specification.php

import gc
import importlib.util
from copy import deepcopy
from pathlib import Path

import numpy as np
import pytest

hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")

testfile_dir = Path(__file__).parent / "data" / "jobinyvon"

testfile_spec_wavelength_path = testfile_dir / "jobinyvon_test_spec.xml"
testfile_spec_wavenumber_path = testfile_dir / "jobinyvon_test_spec_3s_cm-1.xml"
testfile_spec_abs_wavenumber_path = testfile_dir / "jobinyvon_test_spec_3s_abs-cm-1.xml"
testfile_spec_energy_path = testfile_dir / "jobinyvon_test_spec_3s_eV.xml"
testfile_linescan_path = testfile_dir / "jobinyvon_test_linescan.xml"
testfile_map_path = testfile_dir / "jobinyvon_test_map_x3-y2.xml"
testfile_map_rotated_path = testfile_dir / "jobinyvon_test_map_x2-y2_rotated.xml"
testfile_glue_path = testfile_dir / "jobinyvon_test_spec_range.xml"
testfile_spec_count_path = testfile_dir / "jobinyvon_test_spec_3s_counts.xml"


if importlib.util.find_spec("lumispy") is None:
    lumispy_installed = False
else:
    lumispy_installed = True


class TestSpec:
    @classmethod
    def setup_class(cls):
        cls.s = hs.load(
            testfile_spec_wavelength_path,
            reader="JobinYvon",
            use_uniform_signal_axis=True,
        )
        cls.s_non_uniform = hs.load(
            testfile_spec_wavelength_path,
            reader="JobinYvon",
            use_uniform_signal_axis=False,
        )
        cls.s_wn = hs.load(
            testfile_spec_wavenumber_path,
            reader="JobinYvon",
            use_uniform_signal_axis=True,
        )
        cls.s_abs_wn = hs.load(
            testfile_spec_abs_wavenumber_path,
            reader="JobinYvon",
            use_uniform_signal_axis=True,
        )
        cls.s_ev = hs.load(
            testfile_spec_energy_path,
            reader="JobinYvon",
            use_uniform_signal_axis=True,
        )
        cls.s_count = hs.load(
            testfile_spec_count_path,
            reader="JobinYvon",
            use_uniform_signal_axis=True,
        )

    @classmethod
    def teardown_class(cls):
        del cls.s
        del cls.s_non_uniform
        del cls.s_wn
        del cls.s_abs_wn
        del cls.s_ev
        gc.collect()

    @pytest.mark.skipif(lumispy_installed, reason="lumispy is installed")
    def test_signal1D(self):
        hyperspy = pytest.importorskip("hyperspy", reason="hyperspy not installed")
        assert isinstance(self.s, hyperspy._signals.signal1d.Signal1D)

    def test_lumispectrum(self):
        lum = pytest.importorskip("lumispy", reason="lumispy not installed")
        s_lum = hs.load(
            testfile_spec_wavelength_path,
            reader="JobinYvon",
            use_uniform_signal_axis=True,
        )
        assert isinstance(s_lum, lum.signals.luminescence_spectrum.LumiSpectrum)

    def test_intensity_count_unit(self):
        assert self.s_count.metadata.Signal.quantity == "Intensity (Counts)"

    def test_signal_units(self):
        assert self.s_wn.axes_manager.as_dictionary()["axis-0"]["units"] == "1/cm"
        assert self.s_abs_wn.axes_manager.as_dictionary()["axis-0"]["units"] == "1/cm"
        assert self.s_ev.axes_manager.as_dictionary()["axis-0"]["units"] == "eV"

    def test_signal_names(self):
        assert self.s_wn.axes_manager.as_dictionary()["axis-0"]["name"] == "Raman Shift"
        assert (
            self.s_abs_wn.axes_manager.as_dictionary()["axis-0"]["name"] == "Wavenumber"
        )
        assert self.s_ev.axes_manager.as_dictionary()["axis-0"]["name"] == "Energy"

    def test_integration_time(self):
        np.testing.assert_allclose(
            self.s_wn.metadata["Acquisition_instrument"]["Detector"][
                "integration_time"
            ],
            3,
        )
        np.testing.assert_allclose(
            self.s_wn.metadata["Acquisition_instrument"]["Detector"][
                "exposure_per_frame"
            ],
            3,
        )
        np.testing.assert_allclose(
            self.s_wn.metadata["Acquisition_instrument"]["Detector"]["frames"], 1
        )

    def test_data(self):
        spec_data = [
            1496,
            1242,
            1094,
            986,
            948,
            900,
            858,
            855,
            840,
            822,
            824,
            820,
            810,
            809,
            791,
            781,
            771,
            782,
            795,
            790,
            777,
            771,
            769,
            767,
            756,
            755,
            759,
            740,
            743,
            763,
            759,
            727,
            764,
            760,
        ]
        assert spec_data[::-1] == self.s.data.tolist()
        np.testing.assert_allclose(self.s.data, self.s_non_uniform.data)

    def test_axes(self):
        spec_axes = {
            "axis-0": {
                "_type": "UniformDataAxis",
                "name": "Wavelength",
                "units": "nm",
                "navigate": False,
                "is_binned": False,
                "size": 34,
            }
        }

        spec_axes_non_uniform = {
            "axis-0": {
                "_type": "DataAxis",
                "name": "Wavelength",
                "units": "nm",
                "navigate": False,
                "is_binned": False,
            }
        }

        non_uniform_axis_values = np.array(
            [
                537.361,
                536.918,
                536.474,
                536.031,
                535.586,
                535.142,
                534.697,
                534.252,
                533.807,
                533.361,
                532.915,
                532.468,
                532.022,
                531.575,
                531.128,
                530.68,
                530.232,
                529.784,
                529.336,
                528.887,
                528.438,
                527.988,
                527.539,
                527.089,
                526.639,
                526.188,
                525.737,
                525.286,
                524.835,
                524.383,
                523.931,
                523.479,
                523.027,
                522.574,
            ]
        )

        uniform_axis_manager = deepcopy(self.s.axes_manager.as_dictionary())
        non_uniform_axis_manager = deepcopy(
            self.s_non_uniform.axes_manager.as_dictionary()
        )
        np.testing.assert_allclose(
            uniform_axis_manager["axis-0"].pop("scale"), 0.4481, atol=0.0001
        )
        np.testing.assert_allclose(
            uniform_axis_manager["axis-0"].pop("offset"), 522.6, atol=0.05
        )

        np.testing.assert_allclose(
            non_uniform_axis_values[::-1],
            non_uniform_axis_manager["axis-0"].pop("axis"),
        )
        assert spec_axes == uniform_axis_manager
        assert spec_axes_non_uniform == non_uniform_axis_manager

    def test_original_metadata(self):
        spec_original_metadata = {
            "date": {"Acquired": "27.06.2022 16:26:24"},
            "experimental_setup": {
                "Acq. time (s)": 1.0,
                "Accumulations": 2.0,
                "Range": "Visible",
                "Autofocus": "Off",
                "AutoExposure": "Off",
                "Spike filter": "Multiple accum.",
                "Delay time (s)": 0.0,
                "Binning": 30.0,
                "Readout mode": "Signal",
                "DeNoise": "Off",
                "ICS correction": "Off",
                "Dark correction": "Off",
                "Inst. Process": "Off",
                "Detector temperature (°C)": -118.94,
                "Instrument": "LabRAM HR Evol",
                "Detector": "Symphony VIS",
                "Objective": 100.0,
                "Grating (gr/mm)": 1800.0,
                "ND Filter (%)": 10.0,
                "Laser (nm)": 632.817,
                "Spectro (nm)": 530.0006245,
                "Hole": 100.02125,
                "Laser Pol. (°)": 0.0,
                "Raman Pol. (°)": 0.0,
                "StageXY": "Marzhauser",
                "StageZ": "Marzhauser",
                "X (µm)": 0.0,
                "Y (µm)": 0.0,
                "Z (µm)": 0.0,
                "Full time(s)": 3.0,
                "measurement_type": "Spectrum",
                "title": "jobinyvon_test_spec",
                "signal type": "Intens",
                "signal units": "Cnt/sec",
            },
            "file_information": {
                "Project": "A",
                "Sample": "test",
                "Site": "C",
                "Title": "ev21738_1",
                "Remark": "PL",
                "Date": "27.06.2022 16:26",
            },
        }
        assert spec_original_metadata == self.s.original_metadata.as_dictionary()
        assert (
            spec_original_metadata
            == self.s_non_uniform.original_metadata.as_dictionary()
        )

    def test_metadata(self):
        metadata = deepcopy(self.s.metadata.as_dictionary())
        metadata_non_uniform = deepcopy(self.s_non_uniform.metadata.as_dictionary())
        assert (
            metadata_non_uniform["General"]["FileIO"]["0"]["io_plugin"]
            == "rsciio.jobinyvon"
        )
        assert metadata["General"]["FileIO"]["0"]["io_plugin"] == "rsciio.jobinyvon"
        assert metadata["General"]["date"] == "27.06.2022"
        assert metadata["General"]["original_filename"] == str(
            testfile_spec_wavelength_path.name
        )
        assert metadata["General"]["time"] == "16:26:24"
        assert metadata["Signal"]["quantity"] == "Intensity (Counts/s)"
        assert metadata["Signal"]["signal_dimension"] == 1
        np.testing.assert_allclose(
            metadata["Acquisition_instrument"]["Detector"]["binning"], 30
        )
        np.testing.assert_allclose(
            metadata["Acquisition_instrument"]["Detector"]["exposure_per_frame"], 1
        )
        np.testing.assert_allclose(
            metadata["Acquisition_instrument"]["Detector"]["frames"], 2
        )
        np.testing.assert_allclose(
            metadata["Acquisition_instrument"]["Detector"]["integration_time"], 2
        )
        np.testing.assert_allclose(
            metadata["Acquisition_instrument"]["Detector"]["temperature"], -118.94
        )
        np.testing.assert_allclose(
            metadata["Acquisition_instrument"]["Detector"]["delay_time"], 0
        )
        assert metadata["Acquisition_instrument"]["Detector"]["model"] == "Symphony VIS"
        assert (
            metadata["Acquisition_instrument"]["Detector"]["processing"][
                "auto_exposure"
            ]
            == "Off"
        )
        assert (
            metadata["Acquisition_instrument"]["Detector"]["processing"]["autofocus"]
            == "Off"
        )
        assert (
            metadata["Acquisition_instrument"]["Detector"]["processing"][
                "dark_correction"
            ]
            == "Off"
        )
        assert (
            metadata["Acquisition_instrument"]["Detector"]["processing"]["de_noise"]
            == "Off"
        )
        assert (
            metadata["Acquisition_instrument"]["Detector"]["processing"][
                "ics_correction"
            ]
            == "Off"
        )
        assert (
            metadata["Acquisition_instrument"]["Detector"]["processing"]["inst_process"]
            == "Off"
        )
        assert (
            metadata["Acquisition_instrument"]["Detector"]["processing"]["spike_filter"]
            == "Multiple accum."
        )
        np.testing.assert_allclose(
            metadata["Acquisition_instrument"]["Laser"]["objective_magnification"], 100
        )
        np.testing.assert_allclose(
            metadata["Acquisition_instrument"]["Laser"]["wavelength"], 632.817
        )
        np.testing.assert_allclose(
            metadata["Acquisition_instrument"]["Laser"]["Filter"]["optical_density"],
            0.1,
        )
        np.testing.assert_allclose(
            metadata["Acquisition_instrument"]["Spectrometer"]["Grating"][
                "groove_density"
            ],
            1800,
        )
        np.testing.assert_allclose(
            metadata["Acquisition_instrument"]["Spectrometer"]["central_wavelength"],
            530.0006245,
        )
        np.testing.assert_allclose(
            metadata["Acquisition_instrument"]["Spectrometer"]["entrance_slit_width"],
            1.0002125,
        )
        np.testing.assert_allclose(
            metadata["Acquisition_instrument"]["Spectrometer"]["pinhole"],
            1.0002125,
        )
        np.testing.assert_allclose(
            metadata["Acquisition_instrument"]["Spectrometer"]["Polarizer"]["angle"],
            0,
        )
        np.testing.assert_allclose(
            metadata["Acquisition_instrument"]["Laser"]["Polarizer"]["angle"],
            0,
        )
        assert (
            metadata["Acquisition_instrument"]["Spectrometer"]["model"]
            == "LabRAM HR Evol"
        )
        assert (
            metadata["Acquisition_instrument"]["Spectrometer"]["spectral_range"]
            == "Visible"
        )
        assert (
            metadata["Acquisition_instrument"]["Spectrometer"]["Polarizer"][
                "polarizer_type"
            ]
            == "No"
        )
        assert (
            metadata["Acquisition_instrument"]["Laser"]["Polarizer"]["polarizer_type"]
            == "L/2 vis L"
        )

        ## remove FileIO for comparison (timestamp varies)
        ## plugin is already tested for both (above)
        del metadata["General"]["FileIO"]
        del metadata_non_uniform["General"]["FileIO"]

        assert metadata == metadata_non_uniform


class TestLinescan:
    @classmethod
    def setup_class(cls):
        cls.s = hs.load(
            testfile_linescan_path,
            reader="JobinYvon",
            use_uniform_signal_axis=True,
        )
        cls.s_non_uniform = hs.load(
            testfile_linescan_path,
            reader="JobinYvon",
            use_uniform_signal_axis=False,
        )

    @classmethod
    def teardown_class(cls):
        del cls.s
        del cls.s_non_uniform
        gc.collect()

    def test_data(self):
        linescan_row0 = [
            1614,
            1317,
            1140,
            1035,
            970,
            931,
            901,
            868,
            864,
            845,
            843,
            847,
            831,
            834,
            813,
            810,
            807,
            817,
            807,
            798,
            800,
            797,
            788,
            804,
            767,
            778,
            778,
            787,
            775,
            790,
            769,
            778,
            780,
            783,
        ]
        linescan_row1 = [
            1509,
            1251,
            1087,
            1002,
            934,
            896,
            866,
            830,
            837,
            831,
            815,
            792,
            784,
            811,
            796,
            799,
            794,
            784,
            788,
            773,
            780,
            787,
            797,
            779,
            780,
            765,
            770,
            757,
            758,
            758,
            741,
            769,
            765,
            775,
        ]
        linescan_row2 = [
            1546,
            1292,
            1124,
            1020,
            950,
            901,
            890,
            865,
            865,
            847,
            837,
            824,
            827,
            808,
            818,
            809,
            810,
            814,
            798,
            784,
            790,
            785,
            771,
            790,
            786,
            786,
            773,
            771,
            772,
            768,
            782,
            762,
            757,
            781,
        ]
        assert linescan_row0[::-1] == self.s.data.tolist()[0]
        assert linescan_row1[::-1] == self.s.data.tolist()[1]
        assert linescan_row2[::-1] == self.s.data.tolist()[2]
        np.testing.assert_allclose(self.s.data, self.s_non_uniform.data)

    def test_axes(self):
        linescan_axes = {
            "axis-0": {
                "_type": "UniformDataAxis",
                "name": "Y",
                "units": "µm",
                "navigate": True,
                "is_binned": False,
                "size": 3,
                "scale": 0.5,
                "offset": 0.0,
            },
            "axis-1": {
                "_type": "UniformDataAxis",
                "name": "Wavelength",
                "units": "nm",
                "navigate": False,
                "is_binned": False,
                "size": 34,
            },
        }

        linescan_axes_non_uniform = {
            "axis-0": {
                "_type": "UniformDataAxis",
                "name": "Y",
                "units": "µm",
                "navigate": True,
                "is_binned": False,
                "size": 3,
                "scale": 0.5,
                "offset": 0.0,
            },
            "axis-1": {
                "_type": "DataAxis",
                "name": "Wavelength",
                "units": "nm",
                "navigate": False,
                "is_binned": False,
            },
        }

        non_uniform_axis_values = np.array(
            [
                537.361,
                536.918,
                536.474,
                536.031,
                535.586,
                535.142,
                534.697,
                534.252,
                533.807,
                533.361,
                532.915,
                532.468,
                532.022,
                531.575,
                531.128,
                530.68,
                530.232,
                529.784,
                529.336,
                528.887,
                528.438,
                527.988,
                527.539,
                527.089,
                526.639,
                526.188,
                525.737,
                525.286,
                524.835,
                524.383,
                523.931,
                523.479,
                523.027,
                522.574,
            ]
        )

        uniform_axis_manager = deepcopy(self.s.axes_manager.as_dictionary())
        non_uniform_axis_manager = deepcopy(
            self.s_non_uniform.axes_manager.as_dictionary()
        )

        np.testing.assert_allclose(
            uniform_axis_manager["axis-1"].pop("scale"), 0.4481, atol=0.0001
        )
        np.testing.assert_allclose(
            uniform_axis_manager["axis-1"].pop("offset"), 522.6, atol=0.05
        )
        np.testing.assert_allclose(
            non_uniform_axis_values[::-1],
            non_uniform_axis_manager["axis-1"].pop("axis"),
        )
        assert linescan_axes_non_uniform == non_uniform_axis_manager
        assert linescan_axes == uniform_axis_manager


class TestMap:
    @classmethod
    def setup_class(cls):
        cls.s = hs.load(
            testfile_map_path, reader="JobinYvon", use_uniform_signal_axis=True
        )
        cls.s_non_uniform = hs.load(
            testfile_map_path, reader="JobinYvon", use_uniform_signal_axis=False
        )
        cls.s_rotated = hs.load(testfile_map_rotated_path, reader="JobinYvon")

    @classmethod
    def teardown_class(cls):
        del cls.s
        del cls.s_non_uniform
        del cls.s_rotated
        gc.collect()

    def test_rotation_angle(self):
        original_metadata = self.s_rotated.original_metadata.as_dictionary()
        np.testing.assert_allclose(
            original_metadata["experimental_setup"]["rotation angle (rad)"],
            -0.532322511232846,
        )

        metadata = self.s_rotated.metadata.as_dictionary()
        np.testing.assert_allclose(
            metadata["Acquisition_instrument"]["Spectral_image"]["rotation_angle"],
            -30.499833233447433,
        )

        assert (
            metadata["Acquisition_instrument"]["Spectral_image"]["rotation_angle_units"]
            == "°"
        )

    def test_data(self):
        map_row0 = [
            275.5,
            214.5,
            206.5,
            184,
            168,
            171.5,
            158,
            160,
            169.5,
            154.5,
            156.5,
            147,
            152.5,
            159,
            144,
            147.5,
            142,
            155,
            145,
            143,
            149,
            147,
            147,
            152,
            151.5,
            142.5,
            141,
            139.5,
            143,
            143,
            139.5,
            133.5,
            141.5,
            144,
        ]

        map_row1 = [
            295.5,
            234,
            206.5,
            183,
            165,
            164.5,
            145.5,
            143,
            163.5,
            152,
            142,
            146.5,
            146.5,
            142.5,
            145.5,
            138.5,
            137.5,
            132.5,
            147.5,
            138.5,
            140,
            143,
            141,
            131.5,
            142.5,
            132,
            133,
            145.5,
            141,
            132,
            123.5,
            136.5,
            128.5,
            139.5,
        ]
        map_row2 = [
            257.5,
            210,
            198,
            171,
            168,
            168,
            147.5,
            161.5,
            174.5,
            161,
            157.5,
            152.5,
            157,
            144,
            151.5,
            152,
            148,
            149.5,
            154.5,
            153.5,
            155.5,
            149.5,
            144,
            149,
            147,
            153.5,
            146,
            150,
            149.5,
            145.5,
            146,
            140.5,
            133.5,
            136,
        ]
        map_row3 = [
            257,
            202,
            176,
            146,
            143,
            138,
            126,
            128,
            145.5,
            130,
            124.5,
            114,
            122.5,
            119,
            120.5,
            125,
            134.5,
            125.5,
            118,
            126,
            122,
            120,
            120.5,
            125,
            115.5,
            114.5,
            116,
            122.5,
            119.5,
            115,
            121,
            116.5,
            112.5,
            108.5,
        ]
        map_row4 = [
            262,
            206.5,
            183,
            157.5,
            158,
            143.5,
            137,
            135,
            154,
            145.5,
            138.5,
            128.5,
            152,
            152.5,
            160,
            156.5,
            150.5,
            154,
            160.5,
            147,
            152,
            158.5,
            148.5,
            142.5,
            156.5,
            147.5,
            157,
            147,
            149.5,
            146.5,
            144.5,
            142.5,
            135.5,
            136.5,
        ]
        map_row5 = [
            254.5,
            206.5,
            178,
            169,
            148.5,
            139,
            140.5,
            136.5,
            147.5,
            148.5,
            135,
            137.5,
            131.5,
            130,
            120,
            125,
            122.5,
            124,
            124.5,
            133.5,
            129,
            125.5,
            115.5,
            114,
            110,
            120,
            108.5,
            121,
            114,
            116.5,
            115.5,
            119.5,
            108.5,
            108.5,
        ]

        assert map_row0[::-1] == self.s.inav[0, 0].data.tolist()
        assert map_row1[::-1] == self.s.inav[1, 0].data.tolist()
        assert map_row2[::-1] == self.s.inav[2, 0].data.tolist()
        assert map_row3[::-1] == self.s.inav[0, 1].data.tolist()
        assert map_row4[::-1] == self.s.inav[1, 1].data.tolist()
        assert map_row5[::-1] == self.s.inav[2, 1].data.tolist()
        np.testing.assert_allclose(self.s.data, self.s_non_uniform.data)

    def test_axes(self):
        map_axes = {
            "axis-0": {
                "_type": "UniformDataAxis",
                "name": "Y",
                "units": "µm",
                "navigate": True,
                "is_binned": False,
                "size": 2,
                "scale": 2,
                "offset": -1,
            },
            "axis-1": {
                "_type": "UniformDataAxis",
                "name": "X",
                "units": "µm",
                "navigate": True,
                "is_binned": False,
                "size": 3,
                "scale": 2,
                "offset": -2,
            },
            "axis-2": {
                "_type": "UniformDataAxis",
                "name": "Wavelength",
                "units": "nm",
                "navigate": False,
                "is_binned": False,
                "size": 34,
            },
        }

        uniform_axis_manager = deepcopy(self.s.axes_manager.as_dictionary())

        np.testing.assert_allclose(
            uniform_axis_manager["axis-2"].pop("scale"), 1.5416, atol=0.0001
        )
        np.testing.assert_allclose(
            uniform_axis_manager["axis-2"].pop("offset"), 720.967, atol=0.05
        )

        assert map_axes == uniform_axis_manager


class TestGlue:
    @classmethod
    def setup_class(cls):
        cls.s = hs.load(
            testfile_glue_path, reader="JobinYvon", use_uniform_signal_axis=True
        )
        cls.s_non_uniform = hs.load(
            testfile_glue_path, reader="JobinYvon", use_uniform_signal_axis=False
        )

    @classmethod
    def teardown_class(cls):
        del cls.s
        del cls.s_non_uniform
        gc.collect()

    def test_data(self):
        assert np.isclose(self.s.isig[0].data, 238)
        assert np.isclose(self.s.isig[-1].data, 254.265)

    def test_metadata(self):
        original_metadata = self.s.original_metadata.as_dictionary()
        metadata = self.s.metadata.as_dictionary()

        np.testing.assert_allclose(
            original_metadata["experimental_setup"]["Windows"], 4
        )
        np.testing.assert_allclose(
            metadata["Acquisition_instrument"]["Detector"]["glued_spectrum_windows"], 4
        )
        assert metadata["Acquisition_instrument"]["Detector"]["glued_spectrum"] is True



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

import importlib

__all__ = [
    "file",
    "hdf5",
    "path",
    "rgb",
    "tools",
    "xml",
]


def __dir__():
    return sorted(__all__)


def __getattr__(name):
    if name in __all__:
        return importlib.import_module(f".{name}", "rsciio.utils")

    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

import datetime
import logging

import numpy as np
from dateutil import parser, tz

_logger = logging.getLogger(__name__)


def serial_date_to_ISO_format(serial):
    """
    Convert serial_date to a tuple of string (date, time, time_zone) in ISO
    format. By default, the serial date is converted in local time zone.
    """
    dt_utc = serial_date_to_datetime(serial)
    dt_local = dt_utc.astimezone(tz.tzlocal())
    return dt_local.date().isoformat(), dt_local.time().isoformat(), dt_local.tzname()


def ISO_format_to_serial_date(date, time, timezone="UTC"):
    """Convert ISO format to a serial date."""
    if timezone is None or timezone == "Coordinated Universal Time":
        timezone = "UTC"
    dt = parser.parse("%sT%s" % (date, time)).replace(tzinfo=tz.gettz(timezone))
    return datetime_to_serial_date(dt)


def datetime_to_serial_date(dt):
    """Convert datetime.datetime object to a serial date."""
    if dt.tzname() is None:
        dt = dt.replace(tzinfo=tz.tzutc())
    origin = datetime.datetime(1899, 12, 30, tzinfo=tz.tzutc())
    delta = dt - origin
    return float(delta.days) + (float(delta.seconds) / 86400.0)


def serial_date_to_datetime(serial):
    """Convert serial date to a datetime.datetime object."""
    # Excel date&time format
    origin = datetime.datetime(1899, 12, 30, tzinfo=tz.tzutc())
    secs = (serial % 1.0) * 86400
    delta = datetime.timedelta(int(serial), secs)
    return origin + delta


def get_date_time_from_metadata(metadata, formatting="ISO"):
    """
    Get the date and time from a metadata tree.

    Parameters
    ----------
    metadata : metadata dict
    formatting : string, ('ISO', 'datetime', 'datetime64')
        Default: 'ISO'. This parameter set the formatting of the date,
        and the time, it can be ISO 8601 string, datetime.datetime
        or a numpy.datetime64 object. In the later case, the time zone
        is not supported.

    Returns
    -------
        string, datetime.datetime or numpy.datetime64 object

    """
    md_gen = metadata["General"]
    date, time = md_gen.get("date"), md_gen.get("time")
    if date and time:
        dt = parser.parse(f"{date}T{time}")
        time_zone = md_gen.get("time_zone")
        if time_zone:
            dt = dt.replace(tzinfo=tz.gettz(time_zone))
            if dt.tzinfo is None:
                # time_zone metadata must be offset string
                dt = parser.parse(f"{date}T{time}{time_zone}")

    elif not date and time:
        dt = parser.parse(f"{time}").time()
    elif date and not time:
        dt = parser.parse(f"{date}").date()
    else:
        return None

    if formatting == "ISO":
        res = dt.isoformat()
    elif formatting == "datetime":
        res = dt
    # numpy.datetime64 doesn't support time zone
    elif formatting == "datetime64":
        res = np.datetime64(f"{date}T{time}")

    return res


def msfiletime_to_unix(msfiletime):
    """Convert microsoft's filetime to unix datetime used in python
    built-in datetime
    Parameters
    ----------
    msfiletime: 64-bit integer representing number of 10 microsecond ticks
    from 1601

    Returns
    -------
        datetime.datetime object"""
    return datetime.datetime(1601, 1, 1) + datetime.timedelta(
        microseconds=msfiletime / 10
    )



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.


import gc
from pathlib import Path

import dask
import numpy as np
import pytest
from packaging.version import Version

from rsciio.utils._context_manager import dummy_context_manager

hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")
t = pytest.importorskip("traits.api", reason="traits not installed")

from hyperspy.misc.utils import DictionaryTreeBrowser  # noqa: E402

from rsciio.tvips._api import (  # noqa: E402
    TVIPS_RECORDER_FRAME_HEADER,
    TVIPS_RECORDER_GENERAL_HEADER,
    _find_auto_scan_start_stop,
    _get_frame_record_dtype_from_signal,
    _get_main_header_from_signal,
    _guess_image_mode,
    _guess_scan_index_grid,
    _is_valid_first_tvips_file,
    file_reader,
    file_writer,
)

try:
    WindowsError
except NameError:
    WindowsError = None


TEST_DATA_PATH = Path(__file__).parent / "data" / "tvips"


@pytest.fixture()
def fake_metadata_diffraction():
    metadata = {
        "Acquisition_instrument": {
            "TEM": {
                "beam_current": 23,
                "beam_energy": 200,
                "camera_length": 80,
            }
        },
        "General": {
            "date": "1993-06-18",
            "time": "12:34:56",
            "time_zone": "CET",
        },
    }
    return DictionaryTreeBrowser(metadata)


@pytest.fixture()
def fake_metadata_imaging():
    metadata = {
        "Acquisition_instrument": {
            "TEM": {
                "beam_current": 23,
                "beam_energy": 200,
                "magnification": 3000,
            }
        },
        "General": {
            "date": "1993-06-18",
            "time": "12:34:56",
            "time_zone": "CET",
        },
    }
    return DictionaryTreeBrowser(metadata)


@pytest.fixture()
def fake_metadata_confused():
    metadata = {
        "Acquisition_instrument": {
            "TEM": {
                "beam_current": 23,
                "beam_energy": 200,
                "camera_length": 80,
                "magnification": 3000,
            }
        },
        "General": {
            "date": "1993-06-18",
            "time": "12:34:56",
            "time_zone": "CET",
        },
    }
    return DictionaryTreeBrowser(metadata)


@pytest.fixture()
def fake_metadatas(
    fake_metadata_diffraction, fake_metadata_imaging, fake_metadata_confused
):
    return {
        "diffraction": fake_metadata_diffraction,
        "imaging": fake_metadata_imaging,
        "confused": fake_metadata_confused,
    }


@pytest.fixture()
def fake_signal_3D():
    fake_data = np.arange(120).reshape(4, 5, 6).astype(np.uint16)
    fake_signal = hs.signals.Signal2D(fake_data)
    fake_signal.axes_manager[0].scale_as_quantity = "1 nm"
    fake_signal.axes_manager[1].scale_as_quantity = "1 1/nm"
    fake_signal.axes_manager[2].scale_as_quantity = "1 1/nm"
    return fake_signal


@pytest.fixture()
def fake_signal_4D():
    fake_data = np.arange(360).reshape(3, 4, 5, 6).astype(np.uint32)
    fake_signal = hs.signals.Signal2D(fake_data)
    fake_signal.axes_manager[0].scale_as_quantity = "1 nm"
    fake_signal.axes_manager[1].scale_as_quantity = "1 nm"
    fake_signal.axes_manager[2].scale_as_quantity = "1 1/nm"
    fake_signal.axes_manager[3].scale_as_quantity = "1 1/nm"
    return fake_signal


@pytest.fixture()
def fake_signal_5D():
    fake_data = np.arange(720).reshape(2, 3, 4, 5, 6).astype(np.uint64)
    fake_signal = hs.signals.Signal2D(fake_data)
    fake_signal.axes_manager[0].scale_as_quantity = "1 s"
    fake_signal.axes_manager[1].scale_as_quantity = "1 nm"
    fake_signal.axes_manager[2].scale_as_quantity = "1 nm"
    fake_signal.axes_manager[3].scale_as_quantity = "1 1/nm"
    fake_signal.axes_manager[4].scale_as_quantity = "1 1/nm"
    return fake_signal


@pytest.fixture()
def fake_signals(fake_signal_3D, fake_signal_4D, fake_signal_5D):
    return {
        "fake_signal_3D": fake_signal_3D,
        "fake_signal_4D": fake_signal_4D,
        "fake_signal_5D": fake_signal_5D,
    }


@pytest.mark.parametrize(
    "unit, expected_mode",
    [
        ("1/nm", 2),
        ("1/m", 2),
        ("nm", 1),
        ("m", 1),
        (t.Undefined, None),
        ("foo", None),
        ("", None),
        ("s", None),
    ],
)
@pytest.mark.parametrize("sig", ["fake_signal_3D", "fake_signal_4D", "fake_signal_5D"])
def test_guess_image_mode(unit, expected_mode, sig, fake_signals):
    signal = fake_signals[sig]
    signal.axes_manager[-1].units = unit
    mode = _guess_image_mode(signal._to_dictionary())
    assert mode == expected_mode


@pytest.mark.parametrize(
    "unit, expected_scale_factor, version, fheb",
    [
        ("1/pm", 1e3, 2, 0),
        ("um", 1e3, 1, 60),
        ("foo", 1, 2, 12),
    ],
)
@pytest.mark.parametrize("sig", ["fake_signal_3D", "fake_signal_4D", "fake_signal_5D"])
@pytest.mark.parametrize("metadata", ["diffraction", "imaging", "confused", None])
def test_main_header_from_signal(
    unit,
    expected_scale_factor,
    version,
    fheb,
    sig,
    fake_signals,
    metadata,
    fake_metadatas,
):
    signal = fake_signals[sig]
    if metadata is not None:
        signal._metadata = fake_metadatas[metadata]
    signal.axes_manager[-1].units = unit
    signal.axes_manager[-2].units = unit
    original_scale_x = signal.axes_manager[-2].scale
    if unit == "foo":
        cm = pytest.warns(UserWarning)
    else:
        cm = dummy_context_manager()
    with cm:
        header = _get_main_header_from_signal(signal._to_dictionary(), version, fheb)
    assert header["size"] == np.dtype(TVIPS_RECORDER_GENERAL_HEADER).itemsize
    assert header["version"] == version
    assert header["dimx"] == signal.axes_manager[-2].size
    assert header["dimy"] == signal.axes_manager[-1].size
    assert header["offsetx"] == 0
    assert header["offsety"] == 0
    assert header["pixelsize"] == original_scale_x * expected_scale_factor
    assert (
        header["frameheaderbytes"]
        == np.dtype(TVIPS_RECORDER_FRAME_HEADER).itemsize + fheb
    )
    if metadata == "diffraction" and unit == "1/pm":
        assert (
            header["magtotal"]
            == signal.metadata.Acquisition_instrument.TEM.camera_length
        )
    elif metadata == "imaging" and unit == "um":
        assert (
            header["magtotal"]
            == signal.metadata.Acquisition_instrument.TEM.magnification
        )
    else:
        assert header["magtotal"] == 0
    if metadata is None:
        assert header["ht"] == 0
    else:
        assert header["ht"] == signal.metadata.Acquisition_instrument.TEM.beam_energy


@pytest.mark.parametrize("extra_bytes", [0, 20])
@pytest.mark.parametrize("sig", ["fake_signal_3D", "fake_signal_4D", "fake_signal_5D"])
def test_get_frame_record_dtype(sig, fake_signals, extra_bytes):
    signal = fake_signals[sig]
    dt_fh = _get_frame_record_dtype_from_signal(
        signal._to_dictionary(), extra_bytes=extra_bytes
    )
    dimx = signal.axes_manager[-2].size
    dimy = signal.axes_manager[-1].size
    total_size = (
        np.dtype(TVIPS_RECORDER_FRAME_HEADER).itemsize
        + extra_bytes
        + dimx * dimy * signal.data.itemsize
    )
    assert dt_fh.itemsize == total_size


@pytest.mark.parametrize(
    "filename",
    [
        pytest.param("foo_000.bla", marks=pytest.mark.xfail(raises=ValueError)),
        pytest.param("foo_0.tvips", marks=pytest.mark.xfail(raises=ValueError)),
        pytest.param("foo_001.tvips", marks=pytest.mark.xfail(raises=ValueError)),
        pytest.param("foo_0000.TVIPS", marks=pytest.mark.xfail(raises=ValueError)),
        ("foo_000.tvips"),
        ("foo_000.TVIPS"),
    ],
)
def test_valid_tvips_file(filename):
    isvalid = _is_valid_first_tvips_file(filename)
    assert isvalid


@pytest.mark.parametrize(
    "rotators, expected",
    [
        (np.array([0, 0, 1, 2, 3, 4, 5, 6, 0, 0]), (2, 7)),
        (np.array([0, 1, 2, 3, 4, 5, 6, 0, 0]), (1, 6)),
        (np.array([1, 2, 3, 4, 5, 6, 0, 0]), (0, 5)),
        (np.array([0, 0, 1, 2, 3, 4, 5]), (2, 6)),
        (np.array([0, 0, 1, 1, 3, 3, 4, 0]), (2, 6)),
        (np.array([1, 2, 3, 4, 5]), (0, 4)),
        (np.array([0, 0, 0, 0]), (None, None)),
    ],
)
def test_auto_scan_start_stop(rotators, expected):
    start, stop = _find_auto_scan_start_stop(rotators)
    assert start == expected[0]
    assert stop == expected[1]


@pytest.mark.parametrize(
    "rotators, startstop, expected",
    [
        (np.array([0, 0, 1, 2, 3, 4, 5, 6, 0, 0]), None, np.array([2, 3, 4, 5, 6, 7])),
        (np.array([0, 1, 2, 3, 4, 5, 6, 0, 0]), (2, 5), np.array([2, 2, 3, 4, 5])),
        (
            np.array([1, 3, 3, 4, 5, 5, 9, 0, 0]),
            None,
            np.array([0, 0, 1, 3, 4, 5, 5, 5, 6]),
        ),
        (
            np.array([0, 0, 1, 3, 3, 4, 5, 5, 9, 0, 0]),
            None,
            np.array([2, 2, 3, 5, 6, 7, 7, 7, 8]),
        ),
    ],
)
def test_guess_scan_index_grid(rotators, startstop, expected):
    if startstop is None:
        startstop = _find_auto_scan_start_stop(rotators)
    indices = _guess_scan_index_grid(rotators, startstop[0], startstop[1])
    assert np.all(indices == expected)


def _dask_supports_assignment():
    # direct assignment as follows is possible in newer versions (>2021.04.1) of dask,
    # for backward compatibility we use workaround
    return Version(dask.__version__) >= Version("2021.04.1")


@pytest.mark.parametrize(
    "filename, kwargs",
    [
        (
            TEST_DATA_PATH / "test_tvips_2233_000.tvips",
            {
                "scan_shape": "auto",
                "scan_start_frame": 20,
                "hysteresis": 1,
                "rechunking": False,
            },
        ),
        (
            TEST_DATA_PATH / "test_tvips_2345_000.tvips",
            {
                "scan_shape": (2, 3),
                "scan_start_frame": 0,
                "hysteresis": 0,
                "rechunking": "auto",
            },
        ),
        (
            TEST_DATA_PATH / "test_tvips_2345_split_000.tvips",
            {
                "scan_shape": (2, 2),
                "scan_start_frame": 2,
                "hysteresis": -1,
                "rechunking": {0: 1, 1: 1, 2: None, 3: None},
            },
        ),
    ],
)
@pytest.mark.parametrize("wsa", ["x", "y", None])
@pytest.mark.parametrize("lazy", [True, False])
def test_tvips_file_reader(filename, lazy, kwargs, wsa):
    signal = hs.load(filename, lazy=lazy, **kwargs, winding_scan_axis=wsa)
    signal_test = hs.load(filename, lazy=lazy)
    scs = kwargs.get("scan_shape", "auto")
    ssf = kwargs.get("scan_start_frame", 0)
    hyst = kwargs.get("hysteresis", 0)
    sigshape = signal_test.data.shape[-2:]
    if scs == "auto":
        scan_dim = int(np.sqrt(signal_test.data.shape[0]))
        scs = (scan_dim, scan_dim)
        ssf = 0
    signal_test.data = signal_test.data[ssf:].reshape((*scs, *sigshape))

    if not _dask_supports_assignment() and lazy:
        signal_test.compute()
        signal_test.data = signal_test.data.copy()
    if wsa == "x":
        signal_test.data[..., ::2, :, :, :] = signal_test.data[..., ::2, :, :, :][
            ..., :, ::-1, :, :
        ]
        signal_test.data[..., ::2, :, :, :] = np.roll(
            signal_test.data[..., ::2, :, :, :], hyst, axis=-3
        )
    elif wsa == "y":
        signal_test.data[..., ::2, :, :] = signal_test.data[..., ::2, :, :][
            ..., ::-1, :, :, :
        ]
        signal_test.data[..., ::2, :, :] = np.roll(
            signal_test.data[..., ::2, :, :], hyst, axis=-4
        )
    np.testing.assert_allclose(signal_test.data, signal.data, rtol=1e-6)


@pytest.mark.xfail(raises=ValueError)
def test_read_fail_version():
    hs.load(TEST_DATA_PATH / "test_tvips_2345_split_000.tvips", scan_shape="auto")


@pytest.mark.xfail(raises=ValueError)
def test_read_fail_wind_axis():
    hs.load(
        TEST_DATA_PATH / "test_tvips_2345_split_000.tvips",
        scan_shape=(2, 3),
        winding_scan_axis="z",
    )


@pytest.mark.xfail(raises=ValueError)
def test_read_fail_scan_shape():
    hs.load(TEST_DATA_PATH / "test_tvips_2345_split_000.tvips", scan_shape=(3, 3))


@pytest.mark.xfail(raises=ValueError)
def test_write_fail_signal_type(tmp_path):
    fake_signal = hs.signals.BaseSignal(np.zeros((1, 2, 3, 4)))
    fake_signal.save(tmp_path / "test_000.tvips")


@pytest.mark.parametrize(
    "sig, meta, max_file_size, fheb",
    [
        ("fake_signal_3D", "diffraction", None, 0),
        ("fake_signal_4D", "imaging", None, 0),
        ("fake_signal_5D", "diffraction", None, 0),
        ("fake_signal_4D", "diffraction", 100, 0),
        ("fake_signal_5D", "imaging", 100, 20),
        ("fake_signal_5D", None, 700, 10),
    ],
)
@pytest.mark.parametrize("lazy", [True, False])
def test_file_writer(
    sig, meta, max_file_size, fheb, fake_signals, fake_metadatas, lazy, tmp_path
):
    signal = fake_signals[sig]
    if lazy:
        signal = signal.as_lazy()
    if meta is not None:
        metadata = fake_metadatas[meta]
        signal.metadata.add_dictionary(metadata.as_dictionary())
    metadata = signal.metadata

    filepath = tmp_path / "test_tvips_save_000.tvips"
    scan_shape = signal.axes_manager.navigation_shape
    if max_file_size is not None and max_file_size < 500:
        cm = pytest.warns(UserWarning)
    else:
        cm = dummy_context_manager()
    with cm:
        file_writer(
            filepath,
            signal._to_dictionary(),
            max_file_size=max_file_size,
            frame_header_extra_bytes=fheb,
        )
    if max_file_size is None:
        assert len(list(tmp_path.iterdir())) == 1
    else:
        assert len(list(tmp_path.iterdir())) >= 1
    filepath_load = str(tmp_path / "test_tvips_save_000.tvips")
    dtc = file_reader(filepath_load, scan_shape=scan_shape[::-1], lazy=False)[0]
    np.testing.assert_allclose(signal.data, dtc["data"])
    assert signal.data.dtype == dtc["data"].dtype
    if metadata and meta is not None:
        assert dtc["metadata"]["General"]["date"] == metadata.General.date
        assert dtc["metadata"]["General"]["time"] == metadata.General.time
        assert (
            dtc["metadata"]["Acquisition_instrument"]["TEM"]["beam_energy"]
            == metadata.Acquisition_instrument.TEM.beam_energy
        )
        assert (
            dtc["metadata"]["Acquisition_instrument"]["TEM"]["beam_current"]
            == metadata.Acquisition_instrument.TEM.beam_current
        )
    gc.collect()


@pytest.mark.xfail(raises=ValueError)
def test_file_writer_fail(tmp_path):
    signal = hs.signals.Signal1D(np.array([1, 2, 3]))

    file_writer(str(tmp_path / "test.tvips"), signal._to_dictionary())



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

import os
from pathlib import Path

import pytest

from rsciio.utils._tests import expected_is_binned

hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")
imagecodecs = pytest.importorskip(
    "imagecodecs", reason="skipping test_phenom tests, requires imagecodecs"
)

TEST_DATA_PATH = Path(__file__).parent / "data" / "phenom"

ELID2VERSION0 = TEST_DATA_PATH / "Elid2Version0.elid"
ELID2VERSION1 = TEST_DATA_PATH / "Elid2Version1.elid"
ELID2VERSION2 = TEST_DATA_PATH / "Elid2Version2.elid"
ELID2VERSION3 = TEST_DATA_PATH / "Elid2Version3.elid"
ELID2VERSION4 = TEST_DATA_PATH / "Elid2Version4.elid"


@pytest.mark.parametrize(
    ("pathname"),
    [ELID2VERSION0, ELID2VERSION1, ELID2VERSION2, ELID2VERSION3, ELID2VERSION4],
)
def test_elid(pathname):
    s = hs.load(pathname)
    assert len(s) == 11

    assert s[0].data.shape == (16, 20)
    assert s[0].axes_manager.as_dictionary() == {
        "axis-0": {
            "_type": "UniformDataAxis",
            "name": "y",
            "scale": 0.9757792598920122,
            "offset": 0.0,
            "size": 16,
            "units": "µm",
            "navigate": True,
            "is_binned": False,
        },
        "axis-1": {
            "_type": "UniformDataAxis",
            "name": "x",
            "scale": 0.9757792598920122,
            "offset": 0.0,
            "size": 20,
            "units": "µm",
            "navigate": True,
            "is_binned": False,
        },
    }
    assert (
        s[0].metadata["Acquisition_instrument"]["SEM"]["Stage"]["x"]
        == -2.586744298575455
    )
    assert (
        s[0].metadata["Acquisition_instrument"]["SEM"]["Stage"]["y"]
        == -0.7322168400784014
    )
    assert s[0].metadata["Acquisition_instrument"]["SEM"]["beam_energy"] == 15.0
    assert (
        s[0].metadata["Acquisition_instrument"]["SEM"]["microscope"]
        == "MVE027364-0026-L"
    )
    assert s[0].metadata["General"]["date"] == "2019-08-07"
    assert s[0].metadata["General"]["original_filename"] == os.path.split(pathname)[1]
    assert s[0].metadata["General"]["time"] == "09:37:31"
    assert s[0].metadata["General"]["title"] == "Image 1"
    assert s[0].metadata["Signal"]["signal_type"] == ""
    assert s[0].original_metadata["acquisition"]["scan"]["dwellTime"]["value"] == "200"
    assert s[0].original_metadata["acquisition"]["scan"]["dwellTime"]["unit"] == "ns"
    assert (
        s[0].original_metadata["acquisition"]["scan"]["fieldSize"]
        == 0.000019515585197840245
    )
    assert (
        s[0].original_metadata["acquisition"]["scan"]["highVoltage"]["value"] == "-15"
    )
    assert s[0].original_metadata["acquisition"]["scan"]["highVoltage"]["unit"] == "kV"
    assert s[0].original_metadata["pixelWidth"]["value"] == "975.7792598920121"
    assert s[0].original_metadata["pixelWidth"]["unit"] == "nm"
    assert s[0].original_metadata["pixelHeight"]["value"] == "975.7792598920121"
    assert s[0].original_metadata["pixelHeight"]["unit"] == "nm"
    assert s[0].original_metadata["samplePosition"]["x"] == "-0.002586744298575455"
    assert s[0].original_metadata["samplePosition"]["y"] == "-0.0007322168400784014"
    assert s[0].original_metadata["workingDistance"]["value"] == "8.141749999999993"
    assert s[0].original_metadata["workingDistance"]["unit"] == "mm"
    assert (
        s[0].original_metadata["instrument"]["softwareVersion"]
        == "5.4.5.rc1.bb8fbe3.23039"
    )
    assert s[0].original_metadata["instrument"]["type"] == "PhenomXL"
    assert s[0].original_metadata["instrument"]["uniqueID"] == "MVE027364-0026-L"

    assert s[1].metadata["General"]["title"] == "Image 1, Spot 1"
    assert s[1].data.shape == (2048,)
    assert s[1].axes_manager.as_dictionary() == {
        "axis-0": {
            "_type": "UniformDataAxis",
            "name": "Energy",
            "scale": 0.00988676802994421,
            "offset": -0.03634370080990722,
            "size": 2048,
            "units": "keV",
            "navigate": False,
            "is_binned": expected_is_binned(),
        }
    }
    assert s[1].data.tolist()[0:300] == [
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        15,
        16,
        19,
        30,
        52,
        61,
        98,
        125,
        145,
        129,
        114,
        69,
        45,
        22,
        14,
        11,
        18,
        17,
        30,
        26,
        29,
        19,
        16,
        20,
        26,
        29,
        35,
        51,
        59,
        103,
        139,
        157,
        209,
        220,
        179,
        113,
        99,
        65,
        49,
        31,
        36,
        39,
        42,
        35,
        48,
        37,
        55,
        50,
        45,
        46,
        49,
        40,
        49,
        54,
        35,
        49,
        57,
        63,
        71,
        64,
        75,
        76,
        92,
        98,
        83,
        81,
        94,
        118,
        120,
        160,
        215,
        325,
        363,
        368,
        429,
        403,
        376,
        254,
        204,
        173,
        136,
        124,
        102,
        89,
        97,
        84,
        83,
        75,
        83,
        71,
        85,
        101,
        81,
        72,
        87,
        84,
        90,
        93,
        84,
        68,
        93,
        91,
        82,
        86,
        112,
        85,
        84,
        100,
        110,
        118,
        132,
        118,
        125,
        138,
        128,
        135,
        143,
        143,
        136,
        148,
        227,
        301,
        538,
        1077,
        1946,
        3319,
        5108,
        7249,
        9032,
        10755,
        11441,
        10804,
        9219,
        7245,
        5335,
        3568,
        2213,
        1455,
        825,
        543,
        338,
        283,
        196,
        160,
        123,
        104,
        105,
        92,
        88,
        109,
        89,
        88,
        82,
        95,
        88,
        91,
        87,
        108,
        86,
        85,
        59,
        77,
        72,
        58,
        66,
        69,
        64,
        76,
        56,
        67,
        58,
        60,
        59,
        71,
        56,
        57,
        62,
        50,
        67,
        59,
        59,
        52,
        45,
        60,
        53,
        57,
        59,
        39,
        43,
        55,
        54,
        40,
        43,
        37,
        39,
        41,
        52,
        39,
        53,
        41,
        48,
        40,
        41,
        45,
        36,
        45,
        32,
        40,
        44,
        43,
        55,
        50,
        45,
        59,
        45,
        44,
        66,
        52,
        67,
        74,
        83,
        90,
        92,
        114,
        130,
        131,
        114,
        100,
        100,
        106,
        103,
        84,
        87,
        77,
        76,
        82,
        83,
        78,
        81,
        63,
        49,
        54,
        64,
        45,
        41,
        40,
        41,
        38,
        50,
        39,
        45,
        44,
        42,
        44,
        31,
        36,
        38,
        37,
        55,
        40,
        32,
        34,
        32,
        34,
        37,
        27,
        28,
        45,
        35,
        24,
        40,
        22,
        29,
        33,
        33,
        44,
        34,
    ]
    assert (
        s[1].original_metadata["acquisition"]["scan"]["detectors"]["EDS"]["live_time"]
        == 5.7203385
    )
    assert (
        s[1].original_metadata["acquisition"]["scan"]["detectors"]["EDS"]["real_time"]
        == 10.162500000000001
    )
    assert (
        s[1].original_metadata["acquisition"]["scan"]["detectors"]["EDS"][
            "fast_peaking_time"
        ]
        == 100e-9
    )
    assert (
        s[1].original_metadata["acquisition"]["scan"]["detectors"]["EDS"][
            "slow_peaking_time"
        ]
        == 11.2e-6
    )

    assert s[2].metadata["General"]["title"] == "Image 1, Region 2"
    assert s[2].data.shape == (2048,)
    assert s[2].axes_manager.as_dictionary() == {
        "axis-0": {
            "_type": "UniformDataAxis",
            "name": "Energy",
            "scale": 0.00988676802994421,
            "offset": -0.03634370080990722,
            "size": 2048,
            "units": "keV",
            "navigate": False,
            "is_binned": expected_is_binned(),
        }
    }
    assert (
        s[2].original_metadata["acquisition"]["scan"]["detectors"]["EDS"]["live_time"]
        == 6.5802053
    )
    assert (
        s[2].original_metadata["acquisition"]["scan"]["detectors"]["EDS"]["real_time"]
        == 10.177700000000003
    )

    assert s[3].metadata["General"]["title"] == "Image 1, Map 3"
    assert s[3].data.shape == (16, 16, 2048)
    assert s[3].axes_manager.as_dictionary() == {
        "axis-0": {
            "_type": "UniformDataAxis",
            "name": "y",
            "scale": 1.2197240748650153,
            "offset": 0.0,
            "size": 16,
            "units": "µm",
            "navigate": True,
            "is_binned": False,
        },
        "axis-1": {
            "_type": "UniformDataAxis",
            "name": "x",
            "scale": 1.2197240748650153,
            "offset": 0.0,
            "size": 16,
            "units": "µm",
            "navigate": True,
            "is_binned": False,
        },
        "axis-2": {
            "_type": "UniformDataAxis",
            "name": "X-ray energy",
            "scale": 0.00988676802994421,
            "offset": -0.03634370080990722,
            "size": 2048,
            "units": "keV",
            "navigate": False,
            "is_binned": expected_is_binned(),
        },
    }
    assert (
        s[3].original_metadata["acquisition"]["scan"]["detectors"]["EDS"]["live_time"]
        == 4.047052
    )
    assert (
        s[3].original_metadata["acquisition"]["scan"]["detectors"]["EDS"]["real_time"]
        == 3.0005599999999997
    )

    assert s[4].metadata["General"]["title"] == "Image 1, Line 4"
    assert s[4].data.shape == (64, 2048)
    assert s[4].axes_manager.as_dictionary() == {
        "axis-0": {
            "_type": "UniformDataAxis",
            "name": "i",
            "scale": 1.0,
            "offset": 0.0,
            "size": 64,
            "units": "points",
            "navigate": True,
            "is_binned": False,
        },
        "axis-1": {
            "_type": "UniformDataAxis",
            "name": "X-ray energy",
            "scale": 0.00988676802994421,
            "offset": -0.03634370080990722,
            "size": 2048,
            "units": "keV",
            "navigate": False,
            "is_binned": expected_is_binned(),
        },
    }
    assert (
        s[4].original_metadata["acquisition"]["scan"]["detectors"]["EDS"]["live_time"]
        == 5.504343599999998
    )
    assert (
        s[4].original_metadata["acquisition"]["scan"]["detectors"]["EDS"]["real_time"]
        == 6.410299999999996
    )

    assert s[5].metadata["General"]["title"] == "Image 1, Map 6"
    assert s[5].data.shape == (16, 16, 2048)
    assert s[5].axes_manager.as_dictionary() == {
        "axis-0": {
            "_type": "UniformDataAxis",
            "name": "y",
            "scale": 1.2197240748650153,
            "offset": 0.0,
            "size": 16,
            "units": "µm",
            "navigate": True,
            "is_binned": False,
        },
        "axis-1": {
            "_type": "UniformDataAxis",
            "name": "x",
            "scale": 1.2197240748650153,
            "offset": 0.0,
            "size": 16,
            "units": "µm",
            "navigate": True,
            "is_binned": False,
        },
        "axis-2": {
            "_type": "UniformDataAxis",
            "name": "X-ray energy",
            "scale": 0.009886797201840245,
            "offset": -0.04478043655810262,
            "size": 2048,
            "units": "keV",
            "navigate": False,
            "is_binned": expected_is_binned(),
        },
    }
    assert (
        s[5].original_metadata["acquisition"]["scan"]["detectors"]["EDS"]["live_time"]
        == 4.5919591
    )
    assert (
        s[5].original_metadata["acquisition"]["scan"]["detectors"]["EDS"]["real_time"]
        == 3.00056
    )

    assert s[6].metadata["General"]["title"] == "Image 1, Difference 3 - 6"
    assert s[6].data.shape == (2048,)
    assert s[6].axes_manager.as_dictionary() == {
        "axis-0": {
            "_type": "UniformDataAxis",
            "name": "Energy",
            "scale": 0.00988676802994421,
            "offset": -0.03634370080990722,
            "size": 2048,
            "units": "keV",
            "navigate": False,
            "is_binned": expected_is_binned(),
        }
    }

    assert s[7].metadata["General"]["title"] == "385test - spectrum"
    assert s[7].data.shape == (24, 32)
    assert s[7].axes_manager.as_dictionary() == {
        "axis-0": {
            "_type": "UniformDataAxis",
            "name": "y",
            "scale": 1.0,
            "offset": 0.0,
            "size": 24,
            "units": "points",
            "navigate": True,
            "is_binned": False,
        },
        "axis-1": {
            "_type": "UniformDataAxis",
            "name": "x",
            "scale": 1.0,
            "offset": 0.0,
            "size": 32,
            "units": "points",
            "navigate": True,
            "is_binned": False,
        },
    }
    assert "acquisition" not in s[7].original_metadata

    assert s[8].metadata["General"]["title"] == "385test - spectrum, MSA 1"
    assert s[8].data.shape == (2048,)
    assert s[8].axes_manager.as_dictionary() == {
        "axis-0": {
            "_type": "UniformDataAxis",
            "name": "Energy",
            "scale": 0.0098868,
            "offset": -0.0363437,
            "size": 2048,
            "units": "keV",
            "navigate": False,
            "is_binned": expected_is_binned(),
        }
    }
    assert (
        s[8].original_metadata["acquisition"]["scan"]["detectors"]["EDS"]["live_time"]
        == 0.0
    )
    assert (
        s[8].original_metadata["acquisition"]["scan"]["detectors"]["EDS"]["real_time"]
        == 5.066
    )

    assert s[9].metadata["General"]["title"] == "Image 1"
    assert s[9].data.shape == (35, 40)
    assert s[9].axes_manager.as_dictionary() == {
        "axis-0": {
            "_type": "UniformDataAxis",
            "name": "y",
            "scale": 0.8120422280865187,
            "offset": 0.0,
            "size": 35,
            "units": "µm",
            "navigate": True,
            "is_binned": False,
        },
        "axis-1": {
            "_type": "UniformDataAxis",
            "name": "x",
            "scale": 0.8120422280865187,
            "offset": 0.0,
            "size": 40,
            "units": "µm",
            "navigate": True,
            "is_binned": False,
        },
    }
    assert "EDS" not in s[9].original_metadata["acquisition"]["scan"]["detectors"]

    assert s[10].metadata["General"]["title"] == "Image 1, Map 1"
    assert s[10].data.shape == (16, 16, 2048)
    assert s[10].axes_manager.as_dictionary() == {
        "axis-0": {
            "_type": "UniformDataAxis",
            "name": "y",
            "scale": 2.0301055702162967,
            "offset": 0.0,
            "size": 16,
            "units": "µm",
            "navigate": True,
            "is_binned": False,
        },
        "axis-1": {
            "_type": "UniformDataAxis",
            "name": "x",
            "scale": 2.0301055702162967,
            "offset": 0.0,
            "size": 16,
            "units": "µm",
            "navigate": True,
            "is_binned": False,
        },
        "axis-2": {
            "_type": "UniformDataAxis",
            "name": "X-ray energy",
            "scale": 0.009886797201840245,
            "offset": -0.04478043655810262,
            "size": 2048,
            "units": "keV",
            "navigate": False,
            "is_binned": expected_is_binned(),
        },
    }
    assert (
        s[10].original_metadata["acquisition"]["scan"]["detectors"]["EDS"]["live_time"]
        == 4.821238
    )
    assert (
        s[10].original_metadata["acquisition"]["scan"]["detectors"]["EDS"]["real_time"]
        == 3.0005600000000006
    )



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

# The EMD format is a hdf5 standard proposed at Lawrence Berkeley
# National Lab (see https://emdatasets.com/ for more information).
# FEI later developed another EMD format, also based on the hdf5 standard. This
# reader first checked if the file have been saved by Velox (FEI EMD format)
# and use either the EMD class or the FEIEMDReader class to read the file.
# Writing file is only supported for EMD Berkeley file.


import importlib
import json
import logging
import os
import time
from datetime import datetime

import numpy as np
from dateutil import tz

from rsciio.utils._elements import atomic_number2name
from rsciio.utils._units import _UREG, convert_units
from rsciio.utils.hdf5 import (
    _get_keys_from_group,
    _parse_metadata,
    _parse_sub_data_group_metadata,
)

_logger = logging.getLogger(__name__)


def _parse_json(v, encoding="utf-8"):
    return json.loads(v.decode(encoding))


def _get_detector_metadata_dict(om, detector_name):
    detectors_dict = om["Detectors"]
    # find detector dict from the detector_name
    for key in detectors_dict:
        if detectors_dict[key]["DetectorName"] == detector_name:
            return detectors_dict[key]
    return None


PRUNE_WARNING = (
    "No spectrum stream is present in the file and the "
    "spectrum images are saved in a proprietary format, "
    "which is not supported by RosettaSciIO. This is "
    "because it has been 'pruned' or saved a different "
    "software than Velox, e.g. bcf to emd converter. "
    "If you want to open this data don't prune the "
    "file or read bcf file directly (in case the bcf "
    "to emd converter was used)."
)


class FeiEMDReader(object):
    """
    Class for reading FEI electron microscopy datasets.

    The :class:`~.FeiEMDReader` reads EMD files saved by the FEI Velox
    software package.

    Attributes
    ----------
    dictionaries: list
        List of dictionaries which are passed to the file_reader.
    im_type : string
        String specifying whether the data is an image, spectrum or
        spectrum image.

    """

    def __init__(
        self,
        filename=None,
        select_type=None,
        first_frame=0,
        last_frame=None,
        sum_frames=True,
        sum_EDS_detectors=True,
        rebin_energy=1,
        SI_dtype=None,
        load_SI_image_stack=False,
        lazy=False,
    ):
        # TODO: Finish lazy implementation using the `FrameLocationTable`
        # Parallelise streams reading
        self.filename = filename
        self.select_type = select_type
        self.dictionaries = []
        self.first_frame = first_frame
        self.last_frame = last_frame
        self.sum_frames = sum_frames
        self.sum_EDS_detectors = sum_EDS_detectors
        self.rebin_energy = rebin_energy
        self.SI_data_dtype = SI_dtype
        self.load_SI_image_stack = load_SI_image_stack
        self.lazy = lazy
        self.detector_name = None
        self.original_metadata = {}
        # UUID: label mapping
        self._map_label_dict = {}

    def read_file(self, f):
        self.filename = f.filename
        self.version = _parse_json(f["Version"][0])["version"]
        _logger.info(f"EMD file version: {self.version}")
        self.d_grp = f.get("Data")
        self._check_im_type()
        for key in ["Displays", "Operations", "SharedProperties", "Features"]:
            # In Velox emd v11, the "Operation" group is removed:
            # 'operation settings' are moved to \SharedProperties
            # \Features link to \SharedProperties\DataReference
            # in version <11 \Features linked to \Operations
            if key in f.keys():
                self._parse_metadata_group(f.get(key), key)
        if self.im_type == "SpectrumStream":
            self._parse_image_display(f)
        self._read_data(self.select_type)

    def _read_data(self, select_type):
        self.load_images = self.load_SI = self.load_single_spectrum = True
        if select_type == "single_spectrum":
            self.load_images = self.load_SI = False
        elif select_type == "images":
            self.load_SI = self.load_single_spectrum = False
        elif select_type == "spectrum_image":
            self.load_images = self.load_single_spectrum = False
        elif select_type is None:
            pass
        else:
            raise ValueError(
                "`select_type` parameter takes only: `None`, "
                "'single_spectrum', 'images' or 'spectrum_image'."
            )

        if self.im_type == "Image":
            _logger.info("Reading the images.")
            self._read_images()
        elif self.im_type == "Spectrum":
            self._read_single_spectrum()
            self._read_images()
        elif self.im_type == "SpectrumStream":
            self._read_single_spectrum()
            _logger.info("Reading the spectrum image.")
            t0 = time.time()
            self._read_images()
            t1 = time.time()
            if importlib.util.find_spec("sparse") is None:  # pragma: no cover
                raise ModuleNotFoundError(
                    "The 'sparse' library was not found. "
                    "To load EDS spectrum image data from Velox EMD files, "
                    "'sparse' must be installed."
                )
            self._read_spectrum_stream()

            t2 = time.time()
            _logger.info("Time to load images: {} s.".format(t1 - t0))
            _logger.info("Time to load spectrum image: {} s.".format(t2 - t1))

    def _check_im_type(self):
        if "Image" in self.d_grp:
            if "SpectrumImage" in self.d_grp:
                self.im_type = "SpectrumStream"
            else:
                self.im_type = "Image"
        else:
            self.im_type = "Spectrum"

    def _read_single_spectrum(self):
        if not self.load_single_spectrum:
            return
        spectrum_grp = self.d_grp.get("Spectrum")
        if spectrum_grp is None:
            return  # No spectra in the file
        self.detector_name = "EDS"
        for spectrum_sub_group_key in _get_keys_from_group(spectrum_grp):
            self.dictionaries.append(
                self._read_spectrum(spectrum_grp, spectrum_sub_group_key)
            )

    def _read_spectrum(self, spectrum_group, spectrum_sub_group_key):
        spectrum_sub_group = spectrum_group[spectrum_sub_group_key]
        dataset = spectrum_sub_group["Data"]
        if self.lazy:
            import dask.array as da

            data = da.from_array(dataset, chunks=dataset.chunks).T
        else:
            data = dataset[:].T
        original_metadata = _parse_metadata(spectrum_group, spectrum_sub_group_key)
        original_metadata.update(self.original_metadata)

        # Can be used in more recent version of velox emd files
        self.detector_information = self._get_detector_information(original_metadata)

        dispersion, offset, unit = self._get_dispersion_offset(original_metadata)
        axes = []
        if len(data.shape) == 2:
            if data.shape[0] == 1:
                # squeeze
                data = data[0, :]
            else:
                axes = [
                    {
                        "name": "Stack",
                        "offset": 0,
                        "scale": 1,
                        "size": data.shape[0],
                        "navigate": True,
                    }
                ]
        axes.append(
            {
                "name": "Energy",
                "offset": offset,
                "scale": dispersion,
                "size": data.shape[-1],
                "units": "keV",
                "navigate": False,
            },
        )

        md = self._get_metadata_dict(original_metadata)
        md["Signal"]["signal_type"] = "EDS_TEM"

        return {
            "data": data,
            "axes": axes,
            "metadata": md,
            "original_metadata": original_metadata,
            "mapping": self._get_mapping(),
        }

    def _read_images(self):
        # We need to read the image to get the shape of the spectrum image
        if not self.load_images and not self.load_SI:
            return
        # Get the image data group
        image_group = self.d_grp.get("Image")
        if image_group is None:
            return  # No images in the file
        # Get all the subgroup of the image data group and read the image for
        # each of them
        for image_sub_group_key in _get_keys_from_group(image_group):
            image = self._read_image(image_group, image_sub_group_key)
            if not self.load_images:
                # If we don't want to load the images, we stop here
                return
            self.dictionaries.append(image)

    def _read_image(self, image_group, image_sub_group_key):
        """Return a dictionary ready to parse of return to io module"""
        image_sub_group = image_group[image_sub_group_key]
        original_metadata = _parse_metadata(image_group, image_sub_group_key)
        original_metadata.update(self.original_metadata)

        # Can be used in more recent version of velox emd files
        self.detector_information = self._get_detector_information(original_metadata)
        try:
            self.detector_name = self._get_detector_name(image_sub_group_key)
        except KeyError:
            # File version >= 11 doesn't have the "Operations" group anymore
            if self.detector_information is not None:
                self.detector_name = self.detector_information["DetectorName"]

        read_stack = self.load_SI_image_stack or self.im_type == "Image"
        h5data = image_sub_group["Data"]
        # Get the scanning area shape of the SI from the images
        self.spatial_shape = h5data.shape[:-1]
        # For Velox FFT data, dtype must be specified and lazy is not
        # supported due to special dtype. The data is loaded as-is; to get
        # a traditional view the negative half must be created and the data
        # must be re-centered
        # Similar story for DPC signal
        fft_dtype = [
            [("realFloatHalfEven", "<f4"), ("imagFloatHalfEven", "<f4")],
            [("realFloatHalfOdd", "<f4"), ("imagFloatHalfOdd", "<f4")],
        ]
        dpc_dtype = [("realFloat", "<f4"), ("imagFloat", "<f4")]
        if h5data.dtype in fft_dtype or h5data.dtype == dpc_dtype:
            _logger.debug("Found an FFT or DPC, loading as Complex2DSignal")
            real = h5data.dtype.descr[0][0]
            imag = h5data.dtype.descr[1][0]
            if self.lazy:
                import dask.array as da

                data = da.from_array(h5data, chunks=h5data.chunks)
                data = data[real] + 1j * data[imag]
                data = da.transpose(data, axes=[2, 0, 1])
            else:
                data = np.empty(h5data.shape, h5data.dtype)
                h5data.read_direct(data)
                data = data[real] + 1j * data[imag]
                # Set the axes in frame, y, x order
                data = np.rollaxis(data, axis=2)
        else:
            if self.lazy:
                import dask.array as da

                data = da.transpose(
                    da.from_array(h5data, chunks=h5data.chunks), axes=[2, 0, 1]
                )
            else:
                # Workaround for a h5py bug https://github.com/h5py/h5py/issues/977
                # Change back to standard API once issue #977 is fixed.
                # Preallocate the numpy array and use read_direct method, which is
                # much faster in case of chunked data.
                # Do not specify dtype in np.empty, slows down substantially!
                data = np.empty(h5data.shape)
                h5data.read_direct(data)
                # Set the axes in frame, y, x order
                data = np.rollaxis(data, axis=2)

        pix_scale = original_metadata["BinaryResult"].get(
            "PixelSize", {"height": 1.0, "width": 1.0}
        )
        offsets = original_metadata["BinaryResult"].get("Offset", {"x": 0.0, "y": 0.0})
        original_units = original_metadata["BinaryResult"].get("PixelUnitX", "")

        axes = []
        # stack of images
        if not read_stack:
            data = data[0:1, ...]

        if data.shape[0] == 1:
            # Squeeze
            data = data[0, ...]
            i = 0
        else:
            if "FrameTime" in original_metadata["Scan"]:
                frame_time = original_metadata["Scan"]["FrameTime"]
            else:
                _logger.debug("No Frametime found, likely TEM image stack")
                det_ind = original_metadata["BinaryResult"]["DetectorIndex"]
                frame_time = original_metadata["Detectors"][f"Detector-{det_ind}"][
                    "ExposureTime"
                ]
            frame_time, time_unit = self._convert_scale_units(
                frame_time, "s", 2 * data.shape[0]
            )
            axes.append(
                {
                    "index_in_array": 0,
                    "name": "Time",
                    "offset": 0,
                    "scale": frame_time,
                    "size": data.shape[0],
                    "units": time_unit,
                    "navigate": True,
                }
            )
            i = 1
        scale_x, x_unit = self._convert_scale_units(
            pix_scale["width"], original_units, data.shape[i + 1]
        )
        # to avoid mismatching units between x and y axis, use the same unit as x
        # x is chosen as reference, because scalebar used (usually) the horizontal axis
        # and the units conversion is tuned to get decent scale bar
        scale_y = convert_units(float(pix_scale["height"]), original_units, x_unit)
        # Because "axes" only allows one common unit for offset and scale,
        # offset_x, offset_y is converted to the same unit as x_unit
        offset_x = convert_units(float(offsets["x"]), original_units, x_unit)
        offset_y = convert_units(float(offsets["y"]), original_units, x_unit)

        axes.extend(
            [
                {
                    "index_in_array": i,
                    "name": "y",
                    "offset": offset_y,
                    "scale": scale_y,
                    "size": data.shape[i],
                    "units": x_unit,
                    "navigate": False,
                },
                {
                    "index_in_array": i + 1,
                    "name": "x",
                    "offset": offset_x,
                    "scale": scale_x,
                    "size": data.shape[i + 1],
                    "units": x_unit,
                    "navigate": False,
                },
            ]
        )

        md = self._get_metadata_dict(original_metadata)
        if self.detector_name is not None:
            original_metadata["DetectorMetadata"] = _get_detector_metadata_dict(
                original_metadata, self.detector_name
            )
        if image_sub_group_key in self._map_label_dict:
            md["General"]["title"] = self._map_label_dict[image_sub_group_key]

        return {
            "data": data,
            "axes": axes,
            "metadata": md,
            "original_metadata": original_metadata,
            "mapping": self._get_mapping(
                map_selected_element=False, parse_individual_EDS_detector_metadata=False
            ),
        }

    def _get_detector_name(self, key):
        def iDPC_or_dDPC(metadata):
            return "iDPC" if metadata == "true" else "dDPC"

        om = self.original_metadata["Operations"]
        keys = [
            "CameraInputOperation",
            "StemInputOperation",
            "SurfaceReconstructionOperation",
            "MathematicsOperation",
            "DpcOperation",
            "IntegrationOperation",
            "FftOperation",
        ]

        for k in keys:
            if k in om.keys() and k == keys[0]:
                for metadata in om[k].items():
                    # Find the metadata group matching the key in the dataPath
                    if key in metadata[1]["dataPath"]:
                        return metadata[1]["cameraName"]
            if k in om.keys() and k == keys[1]:
                for metadata in om[k].items():
                    # Find the metadata group matching the key in the dataPath
                    if key in metadata[1]["dataPath"]:
                        return metadata[1]["detector"]
            if k in om.keys() and k == keys[2]:
                for metadata in om[k].items():
                    # Look first for the key in the unfilteredDataPath
                    if "unfilteredDataPath" in metadata[1].keys() and (
                        key in metadata[1]["unfilteredDataPath"]
                    ):
                        return iDPC_or_dDPC(metadata[1]["integrationMode"])
                    # Then look for the key in the DataPath
                    if key in metadata[1]["dataPath"]:
                        detector_name = iDPC_or_dDPC(metadata[1]["integrationMode"])
                        if metadata[1]["enableFilter"] == "true":
                            detector_name = "Filtered {}".format(detector_name)
                        return detector_name
            if k in om.keys() and k == keys[3]:
                for metadata in om[k].items():
                    if key in metadata[1]["dataPath"]:
                        if metadata[1]["outputs"][0]["inputIndex"] == "0":
                            return "A-C"
                        elif metadata[1]["outputs"][0]["inputIndex"] == "1":
                            return "B-D"
            if k in om.keys() and k == keys[4]:
                for metadata in om[k].items():
                    if key in metadata[1]["dataPath"]:
                        return "DPC"
            if k in om.keys() and k == keys[5]:
                for metadata in om[k].items():
                    if key in metadata[1]["dataPath"]:
                        return "DCFI"
            if k in om.keys() and k == keys[6]:
                for metadata in om[k].items():
                    if key in metadata[1]["imageOutputPath"]:
                        return "Half FFT"
        return "Unrecognized_image_signal"

    def _get_detector_information(self, om):
        # if the `BinaryResult/Detector` is not available, there should be only
        # one detector in `Detectors`:
        # e.g. original_metadata['Detectors']['Detector-0']
        if "BinaryResult" in om.keys():
            detector_index = om["BinaryResult"].get("DetectorIndex")
        else:
            detector_index = 0
        if detector_index is not None:
            return om["Detectors"]["Detector-{}".format(detector_index)]
        else:
            return None

    def _parse_frame_time(self, original_metadata, factor=1):
        try:
            frame_time = original_metadata["Scan"]["FrameTime"]
            time_unit = "s"
        except KeyError:
            frame_time, time_unit = None, None

        frame_time, time_unit = self._convert_scale_units(frame_time, time_unit, factor)
        return frame_time, time_unit

    def _parse_image_display(self, f):
        if int(self.version) >= 11:
            # - /Displays/ImageDisplay contains the list of all the image displays.
            #   A EDS Map is just an image display.
            # - These entries contain a json encoded dictionary that contains
            #   'data', 'id', 'settings' and 'title'.
            # - The 'id' is the name of the element. 'data' is pointing to the
            #   data reference in SharedProperties/ImageSeriesDataReference/<UUID>
            #   which in turn is pointing to the /Data/Image/<UUID> where the image
            #   data is located.
            om_image_display = self.original_metadata["Displays"]["ImageDisplay"]
            self._map_label_dict = {}
            for v in om_image_display.values():
                if "data" in v.keys():
                    data_key = _parse_json(f.get(v["data"])[0])["dataPath"]
                    self._map_label_dict[data_key.split("/")[-1]] = v["id"]

        else:
            image_display_group = f.get("Presentation/Displays/ImageDisplay")
            key_list = _get_keys_from_group(image_display_group)

            for key in key_list:
                v = _parse_json(image_display_group[key][0])
                data_key = v["dataPath"].split("/")[-1]  # key in data group
                self._map_label_dict[data_key] = v["display"]["label"]

    def _parse_metadata_group(self, group, group_name):
        d = {}
        try:
            for group_key in _get_keys_from_group(group):
                subgroup = group.get(group_key)
                if hasattr(subgroup, "keys"):
                    sub_dict = {}
                    for subgroup_key in _get_keys_from_group(subgroup):
                        v = _parse_json(subgroup[subgroup_key][0])
                        sub_dict[subgroup_key] = v
                else:
                    sub_dict = _parse_json(subgroup[0])
                d[group_key] = sub_dict
        except IndexError:
            _logger.warning("Some metadata can't be read.")
        self.original_metadata.update({group_name: d})

    def _read_spectrum_stream(self):
        if not self.load_SI:
            return
        self.detector_name = "EDS"
        # Try to read the number of frames from Data/SpectrumImage
        try:
            sig = self.d_grp["SpectrumImage"]
            self.number_of_frames = int(
                _parse_json(sig[next(iter(sig))]["SpectrumImageSettings"][0])[
                    "endFramePosition"
                ]
            )
        except Exception:
            _logger.exception(
                "Failed to read the number of frames from Data/SpectrumImage"
            )
            self.number_of_frames = None
        if self.last_frame is None:
            self.last_frame = self.number_of_frames
        elif self.number_of_frames and self.last_frame > self.number_of_frames:
            raise ValueError(
                "The `last_frame` cannot be greater than"
                " the number of frames, %i for this file." % self.number_of_frames
            )

        spectrum_stream_group = self.d_grp.get("SpectrumStream")
        if spectrum_stream_group is None:  # pragma: no cover
            # "Pruned" file, EDS SI data are in the
            # "SpectrumImage" group
            _logger.warning(PRUNE_WARNING)
            return

        subgroup_keys = _get_keys_from_group(spectrum_stream_group)
        if len(subgroup_keys) == 0:
            # "Pruned" file: in Velox emd v11, the "SpectrumStream"
            # group exists but it is empty
            _logger.warning(PRUNE_WARNING)
            return

        def _read_stream(key):
            stream = FeiSpectrumStream(spectrum_stream_group[key], self)
            return stream

        if self.sum_EDS_detectors:
            if len(subgroup_keys) == 1:
                _logger.warning("The file contains only one spectrum stream")
            # Read the first stream
            s0 = _read_stream(subgroup_keys[0])
            streams = [s0]
            # add other stream streams
            if len(subgroup_keys) > 1:
                for key in subgroup_keys[1:]:
                    stream_data = spectrum_stream_group[key]["Data"][:].T[0]
                    if self.lazy:
                        s0.spectrum_image = (
                            s0.spectrum_image
                            + s0.stream_to_sparse_array(stream_data=stream_data)
                        )
                    else:
                        s0.stream_to_array(
                            stream_data=stream_data, spectrum_image=s0.spectrum_image
                        )
        else:
            streams = [_read_stream(key) for key in subgroup_keys]
        if self.lazy:
            for stream in streams:
                sa = stream.spectrum_image.astype(self.SI_data_dtype)
                stream.spectrum_image = sa

        spectrum_image_shape = streams[0].shape
        original_metadata = streams[0].original_metadata
        original_metadata.update(self.original_metadata)

        # Can be used in more recent version of velox emd files
        self.detector_information = self._get_detector_information(original_metadata)

        pixel_size, offsets, original_units = streams[0].get_pixelsize_offset_unit()
        dispersion, offset, unit = self._get_dispersion_offset(original_metadata)

        scale_x, x_unit = self._convert_scale_units(
            pixel_size["width"], original_units, spectrum_image_shape[1]
        )
        # to avoid mismatching units between x and y axis, use the same unit as x
        # x is chosen as reference, because scalebar used (usually) the horizontal axis
        # and the units conversion is tuned to get decent scale bar
        scale_y = convert_units(float(pixel_size["height"]), original_units, x_unit)
        # Because "axes" only allows one common unit for offset and scale,
        # offset_x, offset_y is converted to the same unit as x_unit
        offset_x = convert_units(float(offsets["x"]), original_units, x_unit)
        offset_y = convert_units(float(offsets["y"]), original_units, x_unit)

        i = 0
        axes = []
        # add a supplementary axes when we import all frames individually
        if not self.sum_frames:
            frame_time, time_unit = self._parse_frame_time(
                original_metadata, spectrum_image_shape[i]
            )
            axes.append(
                {
                    "index_in_array": i,
                    "name": "Time",
                    "offset": 0,
                    "scale": frame_time,
                    "size": spectrum_image_shape[i],
                    "units": time_unit,
                    "navigate": True,
                }
            )
            i = 1
        axes.extend(
            [
                {
                    "index_in_array": i,
                    "name": "y",
                    "offset": offset_y,
                    "scale": scale_y,
                    "size": spectrum_image_shape[i],
                    "units": x_unit,
                    "navigate": True,
                },
                {
                    "index_in_array": i + 1,
                    "name": "x",
                    "offset": offset_x,
                    "scale": scale_x,
                    "size": spectrum_image_shape[i + 1],
                    "units": x_unit,
                    "navigate": True,
                },
                {
                    "index_in_array": i + 2,
                    "name": "X-ray energy",
                    "offset": offset,
                    "scale": dispersion,
                    "size": spectrum_image_shape[i + 2],
                    "units": unit,
                    "navigate": False,
                },
            ]
        )

        md = self._get_metadata_dict(original_metadata)
        md["Signal"]["signal_type"] = "EDS_TEM"

        for stream in streams:
            original_metadata = stream.original_metadata
            original_metadata.update(self.original_metadata)
            self.dictionaries.append(
                {
                    "data": stream.spectrum_image,
                    "axes": axes,
                    "metadata": md,
                    "original_metadata": original_metadata,
                    "mapping": self._get_mapping(
                        parse_individual_EDS_detector_metadata=not self.sum_frames
                    ),
                }
            )

    def _get_dispersion_offset(self, original_metadata):
        try:
            for detectorname, detector in original_metadata["Detectors"].items():
                if (
                    original_metadata["BinaryResult"]["Detector"]
                    in detector["DetectorName"]
                ):
                    dispersion = (
                        float(detector["Dispersion"]) / 1000.0 * self.rebin_energy
                    )
                    offset = float(detector["OffsetEnergy"]) / 1000.0
                    return dispersion, offset, "keV"
        except KeyError:
            _logger.warning("The spectrum calibration can't be loaded.")

        return 1, 0, None

    def _convert_scale_units(self, value, units, factor=1):
        if units is None:
            return value, units
        factor /= 2
        v = float(value) * _UREG(units)
        converted_v = (factor * v).to_compact()
        converted_value = float(converted_v.magnitude / factor)
        converted_units = "{:~}".format(converted_v.units)
        return converted_value, converted_units

    def _get_metadata_dict(self, om):
        meta_gen = {}
        meta_gen["original_filename"] = os.path.split(self.filename)[1]
        if self.detector_name is not None:
            meta_gen["title"] = self.detector_name
        # We have only one entry in the original_metadata, so we can't use
        # the mapping of the original_metadata to set the date and time in
        # the metadata: need to set it manually here
        try:
            if "AcquisitionStartDatetime" in om["Acquisition"].keys():
                unix_time = om["Acquisition"]["AcquisitionStartDatetime"]["DateTime"]
            # Workaround when the 'AcquisitionStartDatetime' key is missing
            # This timestamp corresponds to when the data is stored
            elif (
                not isinstance(om["CustomProperties"], str)
                and "Detectors[BM-Ceta].TimeStamp" in om["CustomProperties"].keys()
            ):
                unix_time = (
                    float(
                        om["CustomProperties"]["Detectors[BM-Ceta].TimeStamp"]["value"]
                    )
                    / 1e6
                )
            date, time = self._convert_datetime(unix_time).split("T")
            meta_gen["date"] = date
            meta_gen["time"] = time
            meta_gen["time_zone"] = self._get_local_time_zone()
        except UnboundLocalError:
            # Error seems to come from h5py, covered in the test suite
            # Added in https://github.com/hyperspy/hyperspy/pull/1831
            pass

        meta_sig = {}
        meta_sig["signal_type"] = ""

        return {"General": meta_gen, "Signal": meta_sig}

    def _get_mapping(
        self, map_selected_element=True, parse_individual_EDS_detector_metadata=True
    ):
        mapping = {
            "Acquisition.AcquisitionStartDatetime.DateTime": (
                "General.time_zone",
                lambda x: self._get_local_time_zone(),
            ),
            "Optics.AccelerationVoltage": (
                "Acquisition_instrument.TEM.beam_energy",
                lambda x: float(x) / 1e3,
            ),
            "Optics.CameraLength": (
                "Acquisition_instrument.TEM.camera_length",
                lambda x: float(x) * 1e3,
            ),
            "CustomProperties.StemMagnification.value": (
                "Acquisition_instrument.TEM.magnification",
                float,
            ),
            "Instrument.InstrumentClass": (
                "Acquisition_instrument.TEM.microscope",
                None,
            ),
            "Stage.AlphaTilt": (
                "Acquisition_instrument.TEM.Stage.tilt_alpha",
                lambda x: round(np.degrees(float(x)), 3),
            ),
            "Stage.BetaTilt": (
                "Acquisition_instrument.TEM.Stage.tilt_beta",
                lambda x: round(np.degrees(float(x)), 3),
            ),
            "Stage.Position.x": (
                "Acquisition_instrument.TEM.Stage.x",
                lambda x: round(float(x), 6),
            ),
            "Stage.Position.y": (
                "Acquisition_instrument.TEM.Stage.y",
                lambda x: round(float(x), 6),
            ),
            "Stage.Position.z": (
                "Acquisition_instrument.TEM.Stage.z",
                lambda x: round(float(x), 6),
            ),
            "ImportedDataParameter.Number_of_frames": (
                "Acquisition_instrument.TEM.Detector.EDS.number_of_frames",
                None,
            ),
            "DetectorMetadata.ElevationAngle": (
                "Acquisition_instrument.TEM.Detector.EDS.elevation_angle",
                lambda x: round(float(x), 3),
            ),
            "DetectorMetadata.Gain": (
                "Signal.Noise_properties.Variance_linear_model.gain_factor",
                float,
            ),
            "DetectorMetadata.Offset": (
                "Signal.Noise_properties.Variance_linear_model.gain_offset",
                float,
            ),
        }

        # Parse individual metadata for each EDS detector
        if parse_individual_EDS_detector_metadata:
            mapping.update(
                {
                    "DetectorMetadata.AzimuthAngle": (
                        "Acquisition_instrument.TEM.Detector.EDS.azimuth_angle",
                        lambda x: "{:.3f}".format(np.degrees(float(x))),
                    ),
                    "DetectorMetadata.LiveTime": (
                        "Acquisition_instrument.TEM.Detector.EDS.live_time",
                        lambda x: "{:.6f}".format(float(x)),
                    ),
                    "DetectorMetadata.RealTime": (
                        "Acquisition_instrument.TEM.Detector.EDS.real_time",
                        lambda x: "{:.6f}".format(float(x)),
                    ),
                    "DetectorMetadata.DetectorName": ("General.title", None),
                }
            )

        # Add selected element
        if map_selected_element:
            if int(self.version) >= 11:
                key = "SharedProperties.EDSSpectrumQuantificationSettings"
            else:
                key = "Operations.ImageQuantificationOperation"
            mapping[key] = ("Sample.elements", self._convert_element_list)

        return mapping

    def _convert_element_list(self, d):
        _d = d[d.keys()[0]]
        # "elementSelection" key is missing when no elements are selected for quantification
        if "elementSelection" in _d.keys():
            atomic_number_list = _d["elementSelection"]
            element_list = [
                atomic_number2name[int(atomic_number)]
                for atomic_number in atomic_number_list
            ]
        else:
            element_list = []

        return element_list

    def _convert_datetime(self, unix_time):
        # Since we don't know the actual time zone of where the data have been
        # acquired, we convert the datetime to the local time for convenience
        dt = datetime.fromtimestamp(float(unix_time), tz=tz.tzutc())
        return dt.astimezone(tz.tzlocal()).isoformat().split("+")[0]

    def _get_local_time_zone(self):
        return tz.tzlocal().tzname(datetime.today())


# Below some information we have got from FEI about the format of the stream:
#
# The SI data is stored as a spectrum stream, ‘65535’ means next pixel
# (these markers are also called `Gate pulse`), other numbers mean a spectrum
# count in that bin for that pixel.
# For the size of the spectrum image and dispersion you have to look in
# AcquisitionSettings.
# The spectrum image cube itself stored in a compressed format, that is
# not easy to decode.


class FeiSpectrumStream(object):
    """Read spectrum image stored in FEI's stream format

    Once initialized, the instance of this class supports numpy style
    indexing and slicing of the data stored in the stream format.
    """

    def __init__(self, stream_group, reader):
        self.reader = reader
        self.stream_group = stream_group
        # Parse acquisition settings to get bin_count and dtype
        acquisition_settings_group = stream_group["AcquisitionSettings"]
        acquisition_settings = _parse_json(acquisition_settings_group[0])
        self.bin_count = int(acquisition_settings["bincount"])
        if self.bin_count % self.reader.rebin_energy != 0:
            raise ValueError(
                "The `rebin_energy` needs to be a divisor of the",
                " total number of channels.",
            )
        if self.reader.SI_data_dtype is None:
            self.reader.SI_data_dtype = acquisition_settings["StreamEncoding"]
        # Parse the rest of the metadata for storage
        self.original_metadata = _parse_sub_data_group_metadata(stream_group)
        # If last_frame is None, compute it
        stream_data = self.stream_group["Data"][:].T[0]
        if self.reader.last_frame is None:
            # The information could not be retrieved from metadata
            # we compute, which involves iterating once over the whole stream.
            # This is required to support the `last_frame` feature without
            # duplicating the functions as currently numba does not support
            # parametetrization.
            spatial_shape = self.reader.spatial_shape
            last_frame = int(
                np.ceil(
                    (stream_data == 65535).sum() / (spatial_shape[0] * spatial_shape[1])
                )
            )
            self.reader.last_frame = last_frame
            self.reader.number_of_frames = last_frame
        self.original_metadata["ImportedDataParameter"] = {
            "First_frame": self.reader.first_frame,
            "Last_frame": self.reader.last_frame,
            "Number_of_frames": self.reader.number_of_frames,
            "Rebin_energy": self.reader.rebin_energy,
            "Number_of_channels": self.bin_count,
        }
        # Convert stream to spectrum image
        if self.reader.lazy:
            self.spectrum_image = self.stream_to_sparse_array(stream_data=stream_data)
        else:
            self.spectrum_image = self.stream_to_array(stream_data=stream_data)

    @property
    def shape(self):
        return self.spectrum_image.shape

    def get_pixelsize_offset_unit(self):
        om_br = self.original_metadata["BinaryResult"]
        return om_br["PixelSize"], om_br["Offset"], om_br["PixelUnitX"]

    def stream_to_sparse_array(self, stream_data):
        import rsciio.utils._fei_stream_readers as stream_readers

        """Convert stream in sparse array

        Parameters
        ----------
        stream_data: array

        """
        # Here we load the stream data into memory, which is fine is the
        # arrays are small. We could load them lazily when lazy.
        sparse_array = stream_readers.stream_to_sparse_COO_array(
            stream_data=stream_data,
            spatial_shape=self.reader.spatial_shape,
            first_frame=self.reader.first_frame,
            last_frame=self.reader.last_frame,
            channels=self.bin_count,
            sum_frames=self.reader.sum_frames,
            rebin_energy=self.reader.rebin_energy,
        )
        return sparse_array

    def stream_to_array(self, stream_data, spectrum_image=None):
        """Convert stream to array.

        Parameters
        ----------
        stream_data : numpy.ndarray
        spectrum_image : numpy.ndarray or None
            If array, the data from the stream are added to the array.
            Otherwise it creates a new array and returns it.

        """
        import rsciio.utils._fei_stream_readers as stream_readers

        spectrum_image = stream_readers.stream_to_array(
            stream=stream_data,
            spatial_shape=self.reader.spatial_shape,
            channels=self.bin_count,
            first_frame=self.reader.first_frame,
            last_frame=self.reader.last_frame,
            rebin_energy=self.reader.rebin_energy,
            sum_frames=self.reader.sum_frames,
            spectrum_image=spectrum_image,
            dtype=self.reader.SI_data_dtype,
        )
        return spectrum_image
